#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from typing import Optional, Final, final
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from alive_progress import alive_bar
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


# Number of elements to display in the progress bar
barElems: Final[int] = int(1.E5)


@final
class ProgressBar:
    """ Provide a progress bar outside of the context manager
    """
    __slots__ = ('bar', '_cm', '_len', '_cur', '_chunk', '_pend')

    def __init__(self,
                 title       : Optional[str],
                 value       : int,
                 length      : int  = 33,
                 threshold   : int  = barElems,
                 chunk       : int  = 0,
                 enrich_print: bool = True) -> None:
        # Local imports ----------------------------------------
        from pyhope.common.common import IsInteractive
        # ------------------------------------------------------
        self._len   = value
        self._cur   = 0
        self._chunk = chunk
        self._pend  = 0
        self._cm    = None
        self.bar    = None

        if value > threshold and IsInteractive():
            self._cm = alive_bar(title=title, total=value, length=length, enrich_print=enrich_print)

            # Initialize the progress bar
            self.bar = self._cm.__enter__()

    def step(self, steps: int = 1) -> None:
        # Assume sane calls (steps > 0, no large overshoot); keep logic minimal and fast
        if self.bar is None:
            return None

        if not self._chunk:
            self.bar(steps)
            self._cur += steps
            return None

        # Chunked updates: emit in multiples of _chunk, flush remainder at the end
        self._pend += steps
        self._cur  += steps

        if self._pend >= self._chunk:
            full = (self._pend // self._chunk) * self._chunk
            if full:
                self.bar(full)
                self._pend -= full

        if self._cur >= self._len and self._pend:
            self.bar(self._pend)
            self._pend = 0

    def title(self, title: str) -> None:
        if self.bar is not None:
            self.bar.title(title)

    def close(self) -> None:
        if self._cm is not None:
            _ = self._cm.__exit__(None, None, None)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (C) 2022 Nico Schlömer
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import importlib
from typing import Dict, Final, List, Set, cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def gmsh_to_meshio(gmsh) -> meshio.Mesh:
    """
    Convert a Gmsh object to a meshio object.
    """
    # Local imports ----------------------------------------
    from pyhope.meshio.meshio_ordering import NodeOrdering
    # ------------------------------------------------------

    # Initialize the node ordering
    node_ordering = NodeOrdering()

    # Extract point coords
    idx, points, _ = gmsh.model.mesh.getNodes()
    points  = np.asarray(points).reshape(-1, 3)
    idx    -= 1
    srt     = np.argsort(idx)
    assert np.all(idx[srt] == np.arange(len(idx)))
    points = points[srt]

    # Extract cells
    elem_types, elem_tags, node_tags = gmsh.model.mesh.getElements()
    cells = []
    for elem_type, elem_tags, node_tags in zip(elem_types, elem_tags, node_tags):
        # `elementName', `dim', `order', `numNodes', `localNodeCoord', `numPrimaryNodes'
        num_nodes_per_cell = gmsh.model.mesh.getElementProperties(elem_type)[3]

        node_tags_reshaped = np.asarray(node_tags).reshape(-1, num_nodes_per_cell) - 1
        node_tags_reshaped = node_ordering.ordering_gmsh_to_meshio(elem_type, node_tags_reshaped)

        # NRG: Fix the element ordering
        node_tags_sorted   = node_tags_reshaped[np.argsort(elem_tags)]
        cells.append(meshio.CellBlock(meshio.gmsh.gmsh_to_meshio_type[elem_type], node_tags_sorted))

    cell_sets = {}
    for dim, tag in gmsh.model.getPhysicalGroups():
        # Get offset of the node tags (gmsh sorts elements of all dims in succeeding order of node tags, but order of dims might differ)  # noqa: E501
        elem_types, elem_tags, _ = gmsh.model.mesh.getElements(dim=dim)
        elem_tags_group = {meshio.gmsh.gmsh_to_meshio_type[j]: i for i, j in zip(elem_tags, elem_types)}

        name = gmsh.model.getPhysicalName(dim, tag)
        cell_sets[name] = [[] for _ in range(len(cells))]
        for e in gmsh.model.getEntitiesForPhysicalGroup(dim, tag):
            # elem_types, elem_tags, node_tags
            elem_types, elem_tags, _ = gmsh.model.mesh.getElements(dim, e)
            assert len(elem_types) == len(elem_tags)

            meshio_cell_type = [meshio.gmsh.gmsh_to_meshio_type[type_ele] for type_ele in elem_types]
            # Make sure that the cell type appears only once in the cell list
            idx = []
            for k, cell_block in enumerate(cells):
                if cell_block.type in meshio_cell_type:
                    idx.append(k)

            offset = {meshio_cell_type[j]: np.where(elem_tags_group[meshio_cell_type[j]] == elem_tags[j][0])[0] for j in range(len(idx))}  # noqa: E501
            elem_tags = [offset[j] + np.int64(i - i[0]) for j, i in zip(meshio_cell_type, elem_tags)]

            for j, i in enumerate(idx):
                cell_sets[name][i].append(elem_tags[j])

        cell_sets[name] = [(None if len(idcs) == 0 else np.concatenate(idcs)) for idcs in cell_sets[name]]

    return meshio.Mesh(points, cells, cell_sets=cell_sets)


def meshio_to_gmsh(mesh: meshio.Mesh) -> meshio.Mesh:
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.io.io_gmsh import GMSHCELLTYPES
    # ------------------------------------------------------

    # Instantiate the Gmsh cell type mapping
    gmshCellTypes = GMSHCELLTYPES()

    # Combine volume and surface cells as separate cell blocks
    volume_cells  = [cell_block for cell_block in mesh.cells if cell_block.type in gmshCellTypes.cellTypes3D]
    surface_cells = [cell_block for cell_block in mesh.cells if cell_block.type in gmshCellTypes.cellTypes2D]

    # Build new arrays
    celll:      List[meshio.CellBlock] = []
    celldphys:  List[np.ndarray      ] = []
    celldgeom:  List[np.ndarray      ] = []

    # Unique geometrical entity ids per dimension
    # > 0: 3D entities, 1: 2D entities
    geom_id:    np.ndarray       = np.ones((2,), dtype=int)
    geom_tag:   List[List[int]]  = [[] for _ in range(2)]
    geom_nodes: List[List[int]]  = [[] for _ in range(2)]

    # Set to keep track of nodes already used to represent an entity in gmsh:dim_tags
    usedNodes: Set[int] = set()

    # Process each 3D CellBlock: keep shared connectivity, set tags
    # WARNING: Each 3D CellBlock neets to get its OWN geometrical tag so that
    #          meshio._write_entities() does not see duplicate (dim, tag) across cell blocks
    for cell_block in volume_cells:
        # TODO: Get the physical region from mesh_vars.elems
        zone = 1

        # Shared nodes, so keep the original connectivity
        cb     = meshio.CellBlock(cell_block.type, cell_block.data.copy())
        nCells = len(cell_block.data)
        # Unique 3D geometrical tag for volume block
        tag    = int(geom_id[0])
        phys   = np.full(nCells, int(zone), dtype=int)
        geom   = np.full(nCells, tag,       dtype=int)

        # Choose a representative node for this 3D entity (unique across all entities)
        node_cand = cast(np.ndarray, cb.data).ravel().astype(int)
        node_free = [s for s in node_cand if s not in usedNodes]
        node_used = int(node_free[0] if node_free else node_cand[0])
        if not node_free and node_used in usedNodes:
            # All candidate nodes already used; fall back to the first (still valid, we just can't make it unique)
            # hopout.routine(f'Note: reusing representative node {rep_node} for 3D entity tag {tag}')
            hopout.error('All candidate nodes already used for 3D entity tag {tag}')

        geom_nodes[0].append(node_used)
        usedNodes.add(node_used)

        # Update the lists and next 3D entity ID
        geom_id[ 0] += 1
        geom_tag[0].append(tag)
        celll      .append(cb)
        celldphys  .append(phys)
        celldgeom  .append(geom)

    # Build a reverse mapping from the cell sets to the cell types
    all_cells:      Final[tuple] = tuple(cell_block for cell_block in mesh.cells)
    cellTypeToBC:   Final[dict]  = {s.type: {} for s in all_cells}
    cellTypeToName: Final[list]  = [s.type     for s in all_cells]  # noqa: E272
    BCNameToBCID:   Final[dict]  = {bc.name: bc.bcid for bc in mesh_vars.bcs if bc.bcid is not None}
    del all_cells

    # Find the mapping from surface to BC for all cell sets
    for setID, cellSet in mesh.cell_sets.items():
        for cellID, setCells in enumerate(cellSet):
            if setCells is None:
                continue

            cellName = cellTypeToName[cellID]

            for cell_idx in cast(np.ndarray, setCells):
                BCID = BCNameToBCID.get(setID)
                cellTypeToBC[cellName].update({int(cell_idx): BCID})

    # Process each 2D CellBlock: keep shared point indices, set BCs
    # NOTE: Split by BC id so each physical surface becomes its own entity
    for _, cell_block in enumerate(surface_cells):
        cell_type = cell_block.type
        # NOTE: This will assign 0 to faces not found, assuming they are internal faces
        cellBC = np.asarray([cellTypeToBC[cell_type].get(int(idx), 0) for idx in range(len(cell_block))], dtype=int)

        # Create a separate CellBlock per unique BC id
        for BCID in np.unique(cellBC):
            sel = np.where(cellBC == int(BCID))[0]
            if sel.size == 0:
                continue

            cb     = meshio.CellBlock(cell_type, cell_block.data[sel].copy())
            nCells = len(cell_block.data[sel])
            # Unique 2D geometrical tag for surface block
            tag    = int(geom_id[1])
            phys   = np.full(nCells, int(BCID), dtype=int)
            geom   = np.full(nCells, tag      , dtype=int)

            # Choose a representative node for this 2D entity (unique across all entities)
            node_cand = cast(np.ndarray, cb.data).ravel().astype(int)
            node_free = [s for s in node_cand if s not in usedNodes]
            node_used = int(node_free[0] if node_free else node_cand[0])
            if not node_free and node_used in usedNodes:
                # All candidate nodes already used; fall back to the first (still valid, we just can't make it unique)
                # hopout.routine(f'Note: reusing representative node {rep_node} for 3D entity tag {tag}')
                hopout.error('All candidate nodes already used for 3D entity tag {tag}')

            geom_nodes[1].append(node_used)
            usedNodes.add(node_used)

            # Update lists and next 2D entity ID
            geom_id[ 1] += 1
            geom_tag[1].append(tag)
            celll    .append(cb)
            celldphys.append(phys)
            celldgeom.append(geom)

    # Create new mesh with separate CellBlocks
    gmshMesh = type(mesh)(points = mesh.points,        # noqa: E251
                          cells  = cast(dict, celll))  # noqa: E251

    # Mixed elements require gmsh:physical and gmsh:geometrical
    gmshMesh.cell_data.update({'gmsh:physical'   : cast(list, celldphys)})   # pyright: ignore[reportArgumentType, reportCallIssue]
    gmshMesh.cell_data.update({'gmsh:geometrical': cast(list, celldgeom)})   # pyright: ignore[reportArgumentType, reportCallIssue]

    # Provide entity information for nodes (gmsh:dim_tags)
    # Strategy:
    #  - Default all nodes to the FIRST 3D entity
    #  - Override one unique node per 3D entity with (3, tag) to ensure all 3D entities exist in $Entities
    #  - Override one unique node per 2D entity with (2, tag) to ensure all 2D entities exist in $Entities
    dim_tags = np.zeros((mesh.points.shape[0], 2), dtype=int)
    dim_tags[:, 0] = 3
    dim_tags[:, 1] = int(geom_tag[0][0])

    # Set representatives for ALL 3D entities
    for tag, node_used in zip(geom_tag[0], geom_nodes[0]):
        dim_tags[int(node_used), 0] = 3
        dim_tags[int(node_used), 1] = int(tag)

    # Set representatives for ALL 2D entities
    for tag, node_used in zip(geom_tag[1], geom_nodes[1]):
        dim_tags[int(node_used), 0] = 2
        dim_tags[int(node_used), 1] = int(tag)

    gmshMesh.point_data.update({'gmsh:dim_tags': dim_tags})

    # Add PhysicalNames so groups are not missing in the Gmsh output
    field_data: Dict[str, np.ndarray] = {}

    # Add volume physical group (3D)
    if len(geom_tag[0]) > 0:
        field_data['volume'] = np.array([1, 3], dtype=int)  # id=1, dim=3

    # Add surface physical groups (2D) from boundary conditions
    field_data.update({ name: np.array([int(BCID), 2], dtype=int) for name, BCID in BCNameToBCID.items()
                                                                         if BCID is not None })

    # Update or create field_data attribute
    gmshMesh.field_data = {**getattr(gmshMesh, 'field_data', {}), **field_data}

    return gmshMesh


def MeshioGmshOrderingPatch() -> None:
    """ Monkey-patch MeshIO's Gmsh writer to call NodeOrdering helper
    """

    # Avoid multiple patching
    import meshio.gmsh.common as common
    if getattr(common, '_pyhope_ordering_patched', False):
        return None

    # Lazy import to avoid circulars
    try:
        from pyhope.meshio.meshio_ordering import NodeOrdering
    except Exception:
        return None

    # Patch the common module
    try:
        common._meshio_to_gmsh_order = NodeOrdering().ordering_meshio_to_gmsh
        common._pyhope_ordering_patched = True  # type: ignore[attr-defined]
    except Exception:
        # If assignment fails, bail out
        return

    # Also patch known writer modules that imported the symbol directly at module scope
    # > They reference this wrapper at call time
    for mod_name in ('meshio.gmsh._gmsh', 'meshio.gmsh._gmsh41', 'meshio.gmsh._gmsh22', 'meshio.gmsh._gmsh4'):
        try:
            mod = importlib.import_module(mod_name)
            setattr(mod, '_meshio_to_gmsh_order', NodeOrdering().ordering_meshio_to_gmsh)
        except Exception:
            # If assignment fails, pass
            pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from __future__ import annotations
# import copy
import gc
import sys
from collections import defaultdict
from typing import Final, Optional, Union, cast
# from multiprocessing import Pool
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Typing libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import typing
if typing.TYPE_CHECKING:
    import meshio
    import numpy.typing as npt
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
import pyhope.output.output as hopout
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def flip_analytic(side: int, nbside: np.ndarray) -> int:
    """ Determines the flip of the side-to-side connection based on the analytic side ID
        flip = 1 : 1st node of neighbor side = 1st node of side
        flip = 2 : 2nd node of neighbor side = 1st node of side
        flip = 3 : 3rd node of neighbor side = 1st node of side
        flip = 4 : 4th node of neighbor side = 1st node of side
    """
    # Local imports ----------------------------------------
    from pyhope.common.common import find_index
    # ------------------------------------------------------
    # PERF: Finding in list is faster than numpy
    # return int(np.nonzero(nbside == side)[0])
    return find_index(nbside, side)


def connect_sides(sideIDs: list[int], sides: list, flipID: int) -> None:
    """ Connect the master and slave sides
    """
    # sides[sideIDs[0]].update(
    #     # Master side contains positive global side ID
    #     MS         = 1,                         # noqa: E251
    #     connection = sideIDs[1],                # noqa: E251
    #     flip       = flipID,                    # noqa: E251
    #     nbLocSide  = sides[sideIDs[1]].locSide  # noqa: E251
    # )
    # sides[sideIDs[1]].update(
    #     MS         = 0,                         # noqa: E251
    #     connection = sideIDs[0],                # noqa: E251
    #     flip       = flipID,                    # noqa: E251
    #     nbLocSide  = sides[sideIDs[0]].locSide  # noqa: E251
    # )
    sides[sideIDs[0]].MS         = 1                          # noqa: E251
    sides[sideIDs[0]].connection = sideIDs[1]                 # noqa: E251
    sides[sideIDs[0]].flip       = flipID                     # noqa: E251
    # sides[sideIDs[0]].nbLocSide  = sides[sideIDs[1]].locSide  # noqa: E251
    sides[sideIDs[1]].MS         = 0                          # noqa: E251
    sides[sideIDs[1]].connection = sideIDs[0]                 # noqa: E251
    sides[sideIDs[1]].flip       = flipID                     # noqa: E251
    # sides[sideIDs[1]].nbLocSide  = sides[sideIDs[0]].locSide  # noqa: E251


def find_bc_index(bcs: list, key: str) -> Optional[int]:
    """ Find the index of a BC from its name in the list of BCs
    """
    for iBC, bc in enumerate(bcs):
        if key in bc.name:
            return iBC
        # Try again without the leading 'BC_'
        if key[:3] == 'BC_' and key[3:] in bc.name or \
           key[:3] == 'bc_' and key[3:] in bc.name:
            return iBC
    return None


# def find_closest_side(points: np.ndarray, stree: KDTree, tol: float, msg: str, doMortars: bool = False) -> int:
#     """ Query the tree for the closest side
#     """
#     trSide = stree.query(points)
#
#     # Check if the found side is within tolerance
#     # trSide contains the Euclidean distance and the index of the
#     # opposing side in the nbFaceSet
#     if trSide[0] > tol:
#         # Mortar sides are allowed to be not connected
#         if doMortars:
#             return -1
#
#         hopout.error(f'Could not find {msg} side within tolerance {tol}, exiting...', traceback=True)
#     return cast(int, trSide[1])


# def get_side_id(corners: np.ndarray, side_dict: dict) -> int:
#     """ Get sorted corners and hash them to get the side ID
#     """
#     corners_hash = hash(np.sort(corners).tobytes())
#     return side_dict[corners_hash][0]


def get_nonconnected_sides(sides: list, mesh: meshio.Mesh) -> tuple[list, list[np.ndarray]]:
    """ Get a list of internal sides that are not connected to any
        other side together with a list of their centers
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    # ------------------------------------------------------
    # Update the list
    nConnSide   = [s for s in sides if   s.connection is None  # noqa: E271
                                    and (s.bcid is None or mesh_vars.bcs[s.bcid].type[0] in (0, 1))]

    nConnCenter = [np.mean(mesh.points[s.corners], axis=0) for s in nConnSide]
    return nConnSide, nConnCenter


def periodic_update(sides: tuple, elems: tuple, vv: np.ndarray) -> None:
    """Update the mesh after connecting periodic sides
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.mesh.mesh_common import sidetovol2
    # ------------------------------------------------------
    # Periodic corrections are only supported for hexahedral elements
    if elems[0].type % 100 != 8 or elems[1].type % 100 != 8:
        return

    nGeo:   Final[int]         = mesh_vars.nGeo
    points: Final[npt.NDArray] = mesh_vars.mesh.points
    tol:    Final[float]       = mesh_vars.tolPeriodic

    # for iy, ix in np.ndindex(nodes.shape[:2]):
    #     node   = nodes[ix, iy]
    #     nbNode = nbNodes[indices[ix, iy, 0], indices[ix, iy, 1]]
    #
    #     # Sanity check if the periodic vector matches
    #     if not np.allclose(vv['Dir'], mesh_vars.mesh.points[nbNode] - mesh_vars.mesh.points[node],
    #                        rtol=mesh_vars.tolPeriodic, atol=mesh_vars.tolPeriodic):
    #         hopout.error('Error in periodic update, periodic vector does not match!')
    #
    #     # Center between both points
    #     center = 0.5 * (mesh_vars.mesh.points[node] + mesh_vars.mesh.points[nbNode])
    #
    #     lowerP = copy.copy(center)
    #     upperP = copy.copy(center)
    #     for key, val in enumerate(vv['Dir']):
    #         lowerP[key] -= 0.5 * val
    #         upperP[key] += 0.5 * val
    #
    #     mesh_vars.mesh.points[  node] = lowerP
    #     mesh_vars.mesh.points[nbNode] = upperP

    # Map the meshio nodes to the tensor-product nodes
    elemType = elems[0].type
    nodes    = elems[0].nodes[sidetovol2(nGeo, 0            , sides[0].face, elemType)]
    nbNodes  = elems[1].nodes[sidetovol2(nGeo, sides[1].flip, sides[1].face, elemType)]

    # INFO: THIS CURRENTLY MIGHT NOT WORK SINCE WE POTENTIALLY ONLY HAVE THE CORNER NODES AVAILABLE
    try:
        # Translate to periodic nodes
        nbCheck = np.vectorize(lambda s: mesh_vars.periNodes[(s, mesh_vars.bcs[sides[1].bcid].name)], otypes=[int])(nbNodes)

        # Check if the node IDs match
        if not np.array_equal(nodes, nbCheck):
            # Print the node IDs
            print(hopout.warn(f'NodeIDs side[-]: {nodes  }'))
            print(hopout.warn(f'NodeIDs side[+]: {nbCheck}'))
            hopout.error('Error in periodic update, node IDs do not match!')
    # Fallback to comparison of physical coordinates
    except KeyError:
        # Check if periodic vector matches using vectorized np.allclose
        if not np.allclose(points[nodes] + vv, points[nbNodes], rtol=tol, atol=tol):
            # Print the node coordinates
            print(hopout.warn(f'Coordinates side[-]: {points[nodes]  }'))
            print(hopout.warn(f'Coordinates side[+]: {points[nbNodes]}'))
            hopout.error('Error in periodic update, periodic vector does not match!')

    # Calculate the center for both points
    centers = 0.5 * (points[nodes] + points[nbNodes])

    # Update the mesh points for both node and nbNode
    points[nodes]   = centers - 0.5*vv
    points[nbNodes] = centers + 0.5*vv


# PERF: The parallel version does not really speed up the process
# def init_side_worker(s) -> None:
#     """ Initialize the worker with the side data
#     """
#     global worker_sides
#     worker_sides = s


# PERF: The parallel version does not really speed up the process
# def compute_side_hash(side) -> tuple:
#     """ Compute the hash of the side using the global sides
#     """
#     sorted_bytes = np.sort(worker_sides[side].corners).tobytes()
#     return side, hash(sorted_bytes)


def ConnectMesh() -> None:
    # Local imports ----------------------------------------
    import pyhope.io.io_vars as io_vars
    import pyhope.mesh.mesh_vars as mesh_vars
    from pyhope.common.common_progress import ProgressBar
    # from pyhope.common.common_vars import np_mtp
    from pyhope.io.io_vars import MeshFormat, ELEM, ELEMTYPE
    from pyhope.readintools.readintools import GetLogical
    from pyhope.mesh.connect.connect_mortar import ConnectMortar
    # from pyhope.mesh.mesh_common import sidetovol2
    from pyhope.mesh.mesh_common import face_to_nodes
    # ------------------------------------------------------

    match io_vars.outputformat:
        case MeshFormat.HDF5.value:
            pass
        case _:
            return

    hopout.separator()
    hopout.info('CONNECT MESH...')
    hopout.sep()

    mesh_vars.doPeriodicCorrect = GetLogical('doPeriodicCorrect')
    mesh_vars.doMortars         = GetLogical('doMortars')
    doPeriodicCorrect = mesh_vars.doPeriodicCorrect
    doMortars         = mesh_vars.doMortars

    mesh    = mesh_vars.mesh
    elems   = mesh_vars.elems
    sides   = mesh_vars.sides
    nGeo    = mesh_vars.nGeo

    # Native meshio data
    points: Final[np.ndarray] = mesh.points
    cells:  Final[list]       = mesh.cells
    csets:  Final[dict]       = mesh.cell_sets
    cdict:  Final[dict]       = mesh.cells_dict

    # Set BC and periodic sides
    bcs:    Final[list]       = mesh_vars.bcs
    vvs:    Final[list]       = mesh_vars.vvs

    # Consistency check for 2D boundary conditions
    prefixes: Final[list[str]] = ['quad', 'triangle']
    if not any(k.startswith(p) for p in prefixes for k in cdict.keys()):  # pragma: no cover
        if bcs is not None and len(bcs) > 0:
            print(hopout.warn(f'Detected boundary conditions {[bc.name for bc in bcs]}'))
        hopout.error('Could not find any 2D boundary conditions, exiting...')

    # Use a moderate chunk size to bound intermediate progress updates
    chunk = max(1, min(1000, max(10, int(len(sides)/(400)))))
    bar = ProgressBar(value=len(sides), title='│                 Preparing Sides', length=33, chunk=chunk, threshold=1)

    # Map sides to BC
    # > Create a dict containing only the face corners
    # PERF: The parallel version does not really speed up the process
    # if np_mtp > 0:
    #     # Run in parallel with a chunk size
    #     # > Dispatch the tasks to the workers, minimum 10 tasks per worker, maximum 1000 tasks per worker
    #     sides_lst = tuple(s for elem in elems for s in elem.sides)
    #     with Pool(processes=np_mtp, initializer=init_side_worker, initargs=(sides,)) as pool:
    #         results = pool.map(compute_side_hash, sides_lst)
    #     side_corners = dict(results)
    # else:
    #     side_corners = {side: hash(np.sort(sides[side].corners).tobytes()) for elem in elems for side in elem.sides}
    # Build the reverse dictionary
    corner_side = defaultdict(list)
    for elem in elems:
        for side in cast(Union[list, np.ndarray], elem.sides):
            corner_side[hash(tuple(sorted(sides[side].corners)))].append(side)

    # > Create a dict containing only the periodic corners
    peri_corners = {}

    # Find the mapping to the (N-1)-dim elements
    csetMap = { key: tuple(i for i, cell in enumerate(cset) if cell is not None and cast(np.ndarray, cell).size > 0)
                             for key, cset in csets.items()}

    bar.title('│                Processing Sides')
    for key, cset in csets.items():
        # Check if the set is a BC
        bcID = find_bc_index(bcs, key)

        # Ignore the volume zones
        volumeBC = False
        for iMap in csetMap[key]:
            if not any(s in tuple(cdict)[iMap] for s in ['quad', 'triangle']):
                volumeBC = True
                break
        if volumeBC:
            continue

        if bcID is None:
            hopout.error(f'Could not find BC {key} in list, exiting...')

        # Get the list of sides
        for iMap in csetMap[key]:
            # Cache cell types for this mapping to avoid repeated list creation
            cell_types = tuple(cdict)[iMap]
            # Only 2D faces
            if not any(s in cell_types for s in prefixes):
                continue

            iBCsides = np.array(cset[iMap]).astype(int)
            mapFaces = cells[iMap].data
            # Support for hybrid meshes
            nCorners = 4 if 'quad' in cell_types else 3

            # Map the unique BC sides to our non-unique elem sides
            for iSide in iBCsides:
                # Get the corner nodes
                corners = hash(tuple(sorted(mapFaces[iSide][:nCorners])))

                if corners not in corner_side:
                    print()
                    print(hopout.warn('Malformatted side corners, exiting...'))
                    corners = np.sort(mapFaces[iSide][:nCorners])
                    print(hopout.warn(f'> Side {cell_types}, Nodes {corners}'))
                    for corner in corners:
                        print(hopout.warn('- Coordinates  : [' + ' '.join('{:13.8f}'.format(s) for s in mesh.points[corner]) + ']'))
                    # traceback.print_stack(file=sys.stdout)
                    sys.exit(1)

                # Boundary faces are unique, except for inner/periodic sides
                if len(corner_side[corners]) == 0:
                    continue

                # sideID  = find_keys(face_corners, corners)
                sideIDs = corner_side[corners]
                # Multiple sides with the same corners are only allowed for inner [0,100] and periodic [1] BCs
                match bcs[bcID].type[0]:
                    case 0 | 100:  # Inner side
                        pass
                    case 1:        # Periodic side
                        # Only take the first (positive BC_alpha) side
                        sideIDs = [sideIDs[0]]
                    case _:        # Boundary side
                        # Abort if there are multiple sides with the same corners
                        if len(sideIDs) > 1:
                            hopout.error('Found multiple sides with the same corners, exiting...', traceback=True)

                for sideID in sideIDs:
                    # sides[sideID].update(bcid=bcID)
                    sides[sideID].bcid = bcID

                    # Add the periodic nodes of the periodic sides to the side_corners
                    # > Only negative BC_alpha allowed here
                    if bcs[bcID].type[0] == 1 and bcs[bcID].type[3] > 0:
                        pNodes = hash(tuple(sorted(mesh_vars.periNodes[(s, key)] for s in mapFaces[iSide][:nCorners])))
                        peri_corners[sideID] = pNodes
                        # Update the reverse dictionary immediately
                        corner_side[pNodes].append(sideID)

                    if bcs[bcID].type[0] not in (1, 100):
                        bar.step()

    # Try to connect the inner / periodic sides
    passedTypes = {}
    for val in corner_side.values():
        match len(val):
            case 1:  # BC side
                continue
            case 2:  # Internal side
                sideIDs   = val
                # Flip pyramids
                # if elems[sides[sideIDs[0]].elemID].type % 100 != 5 and \
                #    elems[sides[sideIDs[1]].elemID].type % 100 == 5:
                #     sideIDs   = sideIDs[::-1]

                side0     = sides[sideIDs[0]]
                side1     = sides[sideIDs[1]]
                corners   = side0.corners
                nbcorners = side1.corners

                # Translate to periodic nodes if required
                if side0.bcid is not None and side1.bcid is not None and bcs[side1.bcid].type[0] == 1:
                    nbcorners = np.fromiter((mesh_vars.periNodes[(s, bcs[side1.bcid].name)] for s in side1.corners), dtype=int)

                flipID = flip_analytic(corners[0], nbcorners) + 1

                # Sanity check the flip with the other nodes
                # > INFO: MOVED TO OWN CHECKCONNECT ROUTINE
                # elem   = (elems[side0.elemID], elems[side1.elemID])
                # if elem[0].type % 100 == 8 and elem[1].type % 100 == 8:
                #     # Map the meshio nodes to the tensor-product nodes
                #     elemType = elem[0].type
                #     nodes    = elem[0].nodes[sidetovol2(nGeo, 0     , side0.face, elemType)]
                #     nbNodes  = elem[1].nodes[sidetovol2(nGeo, flipID, side1.face, elemType)]
                #
                #     # Translate to periodic nodes if required
                #     if side0.bcid is not None and side1.bcid is not None and bcs[side1.bcid].type[0] == 1:
                #         nbNodes = np.vectorize(lambda s: mesh_vars.periNodes[(s, bcs[side1.bcid].name)], otypes=[int])(nbNodes)
                #
                #     # Check if the node IDs match
                #     if not np.array_equal(nodes, nbNodes):
                #         # Print the node IDs
                #         print(hopout.warn(f'NodeIDs side[-]: {nodes  }'))
                #         print(hopout.warn(f'NodeIDs side[+]: {nbNodes}'))
                #         hopout.error('Error in connectivity check, node IDs do not match!')

                # Connect the sides
                connect_sides(sideIDs, sides, flipID)
                # Update the progress bar
                bar.step(2)

                # Use guard clauses to avoid unnecessary checks
                if side0.bcid is None:
                    continue     # No boundary condition on first side
                if bcs[side0.bcid].type[0] != 1:
                    continue     # Not a periodic BC on first side
                if bcs[side1.bcid].type[0] != 1:
                    hopout.error('Found internal side with inconsistent BC types, exiting...')
                if not doPeriodicCorrect:
                    continue     # Periodic correction not enabled

                # At this point, we know both sides have periodic BCs
                iVV = bcs[side0.bcid].type[3]
                VV  = vvs[np.abs(iVV) - 1]['Dir'] * np.sign(iVV)
                locSides = tuple(sides[s]        for s in sideIDs)  # noqa: E272
                locElems = tuple(elems[s.elemID] for s in locSides)

                # Only update hexahedral elements
                if any(e.type % 100 != 8 for e in locElems):
                    for e in locElems:
                        passedTypes[e.type] = passedTypes.get(e.type, 0) + 1
                else:
                    periodic_update(locSides, locElems, VV)

            case _:  # Zero or more than 2 sides
                hopout.error('Found internal side with more than two adjacent elements, exiting...', traceback=True)

    if passedTypes:
        print(hopout.warn(hopout.Colors.WARN + '─'*(46-16) + hopout.Colors.END))
        print(hopout.warn('Periodic correction skipped for elements:'))
        print(hopout.warn(hopout.Colors.WARN + '─'*(46-16) + hopout.Colors.END))
        for elemType in ELEM.TYPES:
            if elemType in passedTypes and passedTypes[elemType] > 0:
                print(hopout.warn( ELEMTYPE(elemType) + ': {:12d}'.format(passedTypes[elemType])))

    nConnSide, nConnCenter = get_nonconnected_sides(sides, mesh)

    # Mortar sides
    if doMortars:
        # Mortar connections are not supported between mismatching side types
        if any(len(s.corners) != len(nConnSide[0].corners) for s in nConnSide):
            hopout.error('Mortar connections are not supported between mixed side types, exiting...')

        # Connect the mortar sides
        elems, sides = ConnectMortar(nConnSide, nConnCenter, elems, sides, bar)

    nConnSide, nConnCenter = get_nonconnected_sides(sides, mesh)
    if len(nConnSide) > 0:
        print()  # Empty line for spacing
        for side in nConnSide:
            print(hopout.warn(f'> Element {side.elemID+1}, Side {side.face}, Side {side.sideID+1}'))  # noqa: E501
            elem  = elems[side.elemID]
            # nodes = elem.nodes[sidetovol2(nGeo, 0     , side.face, elem.type)]
            nodes = np.transpose(np.array([elem.nodes[s] for s in face_to_nodes(side.face, elem.type, nGeo)]))
            if elem.type % 100 == 8:
                nodes = np.transpose(points[nodes]         , axes=(2, 0, 1))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:13.8f}'.format(s) for s in nodes[:,  0,  0]) + ']'))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:13.8f}'.format(s) for s in nodes[:,  0, -1]) + ']'))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:13.8f}'.format(s) for s in nodes[:, -1,  0]) + ']'))
                print(hopout.warn('- Coordinates  : [' + ' '.join('{:13.8f}'.format(s) for s in nodes[:, -1, -1]) + ']'))
                if side is not nConnSide[-1]:
                    print()  # Empty line for spacing
            else:
                nodes = points[nodes]
                for node in nodes:
                    print(hopout.warn('- Coordinates  : [' + ' '.join('{:13.8f}'.format(s) for s in node) + ']'))
                if side is not nConnSide[-1]:
                    print()  # Empty line for spacing
        hopout.error('Could not connect {} / {} side{}'.format(len(nConnSide), len(sides), '' if len(sides) == 1 else 's'))

    # Close the progress bar
    bar.close()

    # Run garbage collector to release memory
    gc.collect()

    # Count the sides
    nsides             = len(sides)
    sides_conn         = np.empty(nsides, dtype=bool)
    sides_bc           = np.empty(nsides, dtype=bool)
    sides_periodic     = np.empty(nsides, dtype=bool)
    sides_mortar_big   = np.empty(nsides, dtype=bool)
    sides_mortar_small = np.empty(nsides, dtype=bool)

    for i, s in enumerate(sides):
        sides_conn[        i] = s.connection is not None
        sides_bc[          i] = s.bcid       is not None and bcs[s.bcid].type[0] != 100  # noqa: E272
        sides_periodic[    i] = s.bcid       is not None and bcs[s.bcid].type[0] == 1    # noqa: E272
        sides_mortar_big[  i] = s.connection is not None and s.connection < 0
        sides_mortar_small[i] = s.locMortar  is not None                                 # noqa: E272

    # Count each type of side
    ninnersides        = np.sum( sides_conn & ~sides_bc       & ~sides_mortar_small & ~sides_mortar_big)
    nperiodicsides     = np.sum( sides_conn &  sides_periodic & ~sides_mortar_small & ~sides_mortar_big)
    nbcsides           = np.sum(~sides_conn &  sides_bc       & ~sides_mortar_small & ~sides_mortar_big)
    nmortarbigsides    = np.sum(                                                       sides_mortar_big)
    nmortarsmallsides  = np.sum(                                 sides_mortar_small                    )
    nsides             = len(sides) - nmortarsmallsides

    hopout.sep()
    hopout.info(' Number of sides                : {:12d}'.format(nsides))
    hopout.info(' Number of inner sides          : {:12d}'.format(ninnersides))
    hopout.info(' Number of mortar sides (big)   : {:12d}'.format(nmortarbigsides))
    hopout.info(' Number of mortar sides (small) : {:12d}'.format(nmortarsmallsides))
    hopout.info(' Number of boundary sides       : {:12d}'.format(nbcsides))
    hopout.info(' Number of periodic sides       : {:12d}'.format(nperiodicsides))

    mesh_vars.sides = sides
    mesh_vars.elems = elems

    # hopout.sep()
    # hopout.info('CONNECT MESH DONE!')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (C) 2022 Nico Schlömer
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from dataclasses import dataclass, field
from functools import cache
from typing import Dict, List, Union, Optional
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# from pyhope.mesh.mesh_common import NDOFS_ELEM
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


@cache
def HEXREORDER(order: int, incomplete: Optional[bool] = False) -> tuple[int]:
    """ Converts node ordering from gmsh to meshio format
    """
    EDGEMAP   = (  0,  3,  5,  1,  8, 10, 11,  9,  2,  4,  6,  7)
    FACEMAP   = (  2,  3,  1,  4,  0,  5)

    order    += 1
    nNodes    = 8 + 12*(order - 2) if incomplete else order**3
    map: List = [None for _ in range(nNodes)]

    count = 0
    # Recursively build the mapping
    for iOrder in range(np.floor(order/2).astype(int)):
        # Vertices
        map[count:count+8] = list(range(count, count+8))
        count += 8

        pNodes = (order-2*(iOrder+1))

        # Edges
        for iEdge in range(12):
            iSlice = slice(count + pNodes   *iEdge                , count + pNodes    *(iEdge+1))
            map[iSlice] = [count + pNodes   *(EDGEMAP[iEdge])+iNode for iNode in range(pNodes   )]
        count += pNodes*12

        # Only vertices and edges of the outermost shell required for incomplete elements
        if incomplete:
            return tuple(map)

        # Faces
        for iFace in range(6):
            iSlice = slice(count + pNodes**2*iFace                , count + pNodes**2*(iFace+1))
            map[iSlice] = [count + pNodes**2*(FACEMAP[iFace])+iNode for iNode in range(pNodes**2)]
        count += pNodes**2*6

    if order % 2 != 0:
        map[count] = count

    return tuple(map)


# tet order 3
#
#              2
#            ,/|`\
#          ,8  |  `7              E = order - 1
#        ,/    13   `\            C = 4 + 6*E
#      ,9    16 |     `6          F = ((order - 1)*(order - 2))/2
#    ,/         |       `\        N = total number of vertices
#   0-----4-----'.--5-----1
#    `\.   18    |  19  ,/        Interior vertex numbers
#       11.     12    ,15           for edge 0 <= i <= 5: 4+i*E to 4+(i+1)*E-1
#          `\.   '. 14              for face 0 <= j <= 3: C+j*F to C+(j+1)*F-1
#             10\.|/        in volume           : C+4*F to N-1
#      17        `3
#
# tet order 4
#
#              2
#            ,/|`\
#          10  |  `9              E = order - 1
#        11    18   `8            C = 4 + 6*E
#      12  23   |     `7          F = ((order - 1)*(order - 2))/2
#    ,/  22 24  17      `\        N = total number of vertices
#   0-----4----5'.--6-----1
#    `\.  28-30  | 22-24,21       Interior vertex numbers
#       15.  29 16    ,20           for edge 0 <= i <= 5: 4+i*E to 4+(i+1)*E-1
#          14.   '. 19              for face 0 <= j <= 3: C+j*F to C+(j+1)*F-1
#             13\.|/        in volume           : C+4*F to N-1
#    25-26       `3
#      27
#
# [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 14, 15, 12, 13, 18, 19, 17, 16]

# @cache
# def TETREORDER(order: int, incomplete: Optional[bool] = False) -> tuple[int]:
#     """ Converts node ordering from gmsh to meshio format
#     """
#     # gmsh MTetrahedron.h: static const int e[6][2] = {{0, 1}, {1, 2}, {2, 0}, {3, 0}, {3, 2}, {3, 1}};
#     EDGEMAP   = (  0,  1,  2,  3,  5,  4)
#     # gmsh MTetrahedron.h: static const int f[4][3] = {{0, 2, 1}, {0, 1, 3}, {0, 3, 2}, {3, 1, 2}};
#     FACEMAP   = (  1,  3,  2,  0)
#
#     order    += 1
#     nNodes    = 4 + 6*(order - 1) if incomplete else NDOFS_ELEM(104, order-1)
#     map: List = [None for _ in range(nNodes)]
#
#     count = 0
#     map[count:count+4] = list(range(count, count+4))
#     count += 4
#
#     # Edges
#     pNodes = order - 2
#     for iEdge in range(6):
#         iSlice = slice(count + pNodes   *iEdge                , count + pNodes    *(iEdge+1))
#         if iEdge < 3:
#             map[iSlice] = [count + pNodes   *(EDGEMAP[iEdge])+iNode for iNode in range(pNodes)]
#         else:
#             map[iSlice] = [count + pNodes   *(EDGEMAP[iEdge])+iNode for iNode in reversed(range(pNodes))]
#     count += pNodes*6
#
#     # Only vertices and edges of the outermost shell required for incomplete elements
#     if incomplete:
#         return tuple(map)
#
#     # Faces: FIXME for NGeo>=4
#     if order > 3:
#         fNodes = int(((order - 2)*(order - 3))/2)
#         for iFace in range(4):
#             iSlice = slice(count + fNodes*iFace                , count + fNodes*(iFace+1))
#             map[iSlice] = [count + fNodes*(FACEMAP[iFace])+iNode for iNode in range(fNodes)]
#         count += fNodes*4
#
#     # Inner sides
#     if order > 4:
#         iSlice = slice(count                 , nNodes)
#         map[iSlice] = [i for i in range(count, nNodes)]
#
#     return tuple(map)


@dataclass
class NodeOrdering:
    """
    A dataclass that stores the converstion between the node ordering of meshIO and Gmsh.
    """
    # Dictionary for translation of  meshio types to gmsh codes
    # http://gmsh.info//doc/texinfo/gmsh.html#MSH-file-format-version-2
    _gmsh_typing: Dict[int, str] = field(
            default_factory=lambda: { 1  : 'line'          , 2  : 'triangle'      , 3  : 'quad'          , 4  : 'tetra'         ,
                                      5  : 'hexahedron'    , 6  : 'wedge'         , 7  : 'pyramid'       , 8  : 'line3'         ,
                                      9  : 'triangle6'     , 10 : 'quad9'         , 11 : 'tetra10'       , 12 : 'hexahedron27'  ,
                                      13 : 'wedge18'       , 14 : 'pyramid14'     , 15 : 'vertex'        , 16 : 'quad8'         ,
                                      17 : 'hexahedron20'  , 18 : 'wedge15'       , 19 : 'pyramid13'     , 21 : 'triangle10'    ,
                                      23 : 'triangle15'    , 25 : 'triangle21'    , 26 : 'line4'         , 27 : 'line5'         ,
                                      28 : 'line6'         , 29 : 'tetra20'       , 30 : 'tetra35'       , 31 : 'tetra56'       ,
                                      36 : 'quad16'        , 37 : 'quad25'        , 38 : 'quad36'        , 42 : 'triangle28'    ,
                                      43 : 'triangle36'    , 44 : 'triangle45'    , 45 : 'triangle55'    , 46 : 'triangle66'    ,
                                      47 : 'quad49'        , 48 : 'quad64'        , 49 : 'quad81'        , 50 : 'quad100'       ,
                                      51 : 'quad121'       , 62 : 'line7'         , 63: 'line8'          , 64 : 'line9'         ,
                                      65 : 'line10'        , 66 : 'line11'        , 71 : 'tetra84'       , 72 : 'tetra120'      ,
                                      73 : 'tetra165'      , 74 : 'tetra220'      , 75 : 'tetra286'      , 90 : 'wedge40'       ,
                                      91 : 'wedge75'       , 92 : 'hexahedron64'  , 93 : 'hexahedron125' , 94 : 'hexahedron216' ,
                                      95 : 'hexahedron343' , 96 : 'hexahedron512' , 97 : 'hexahedron729' , 98 : 'hexahedron1000',
                                      106: 'wedge126'      , 107: 'wedge196'      , 108: 'wedge288'      , 109: 'wedge405'      ,
                                      110: 'wedge550'
                                    }
    )

    # Dictionary for conversion Gmsh to meshIO
    # > TODO: IMPLEMENT RECURSIVE MAPPING USING IO_MESHIO/IO_GMSH
    _meshio_ordering: Dict[str, List[int]] = field(
            default_factory=lambda: {  # 0D elements
                                       # > Vertex
                                       # 'vertex'      : [ 0 ],
                                       # 1D elements
                                       # > Line
                                       # 'line'        : [ 0, 1 ],
                                       # 2D elements
                                       # > Triangle
                                       # 'triangle'    : [ 0, 1, 2 ],
                                       # > Quadrilateral
                                       # 'quad'        : [ 0, 1, 2, 3 ],
                                       # 3D elements
                                       # > Tetrahedron
                                       'tetra'       : [ 0, 1, 2, 3 ],
                                       'tetra10'     : [ 0, 1, 2, 3, 4, 5, 6, 7, 9, 8 ],
                                       'tetra20'     : [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 10, 15, 14, 13, 12, 17, 19, 18, 16 ],
                                       'tetra35'     : [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 15, 14, 13, 21, 20, 19, 18, 17, 16,
                                                         25, 26, 27, 32, 33, 31, 28, 30, 29, 22, 24, 23, 34],
                                       # > Wedge
                                       'wedge'       : [ 0, 1, 2, 3, 4, 5],
                                       'wedge15'     : [ 0, 1, 2, 3, 4, 5, 6, 9, 7, 12, 14, 13, 8, 10, 11 ],
                                       # http://davis.lbl.gov/Manuals/VTK-4.5/classvtkQuadraticWedge.html and https://gmsh.info/doc/texinfo/gmsh.html#Node-ordering
                                       'wedge18'     : [ 0, 1, 2, 3, 4, 5, 6, 9, 7, 12, 14, 13, 8, 10, 11, 15, 17, 16 ],
                                       # > Pyramid
                                       'pyramid'     : [ 0, 1, 2, 3, 4],
                                       'pyramid13'   : [ 0, 1, 2, 3, 4, 5, 8, 10, 6, 7, 9, 11, 12 ],
                                       # > Hexahedron: for all hexahedron, we now use analytics
                                    }
    )

    # # Dictionary for conversion meshIO to Gmsh
    # # > TODO: IMPLEMENT RECURSIVE MAPPING USING IO_MESHIO/IO_GMSH
    # _gmsh_ordering: Dict[str, List[int]] = field(
    #         default_factory=lambda: { 'tetra10'     : [ 0, 1, 2, 3, 4, 5, 6, 7, 9, 8 ],
    #                                   'hexahedron20': [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 16, 9, 17, 10, 18, 19, 12, 15, 13, 14 ],
    #                                   'hexahedron27': [ 0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 16, 9, 17, 10, 18, 19, 12, 15, 13, 14, 24,
    #                                                     22, 20, 21, 23, 25, 26 ],
    #                                   'wedge15'     : [ 0, 1, 2, 3, 4, 5, 6, 8, 12, 7, 13, 14, 9, 11, 10 ],
    #                                   'pyramid13'   : [ 0, 1, 2, 3, 4, 5, 8, 9, 6, 10, 7, 11, 12 ],
    #                                 }
    # )

    # Dictionary for translation of  gambit types to gmsh codes
    _gambit_typing: Dict[int, str] = field(
            default_factory=lambda: { 1  : 'line'          , 2  : 'quad'          , 3  : 'triangle'      , 4  : 'hexahedron'    ,
                                      5  : 'wedge'         , 6  : 'tetrahedron'   , 7  : 'pyramid'                              ,
                                    }
    )

    # Dictionary for conversion of Gambit to meshIO
    _gambit_ordering: Dict[str, List[int]] = field(
            default_factory=lambda: {  # 0D elements
                                       # 1D elements
                                       # 2D elements
                                       # 3D elements
                                       # > Hexahedron
                                       'hexahedron':   [0, 1, 3, 2, 4, 5, 7, 6],
                                    }
    )

    def ordering_gmsh_to_meshio(self, elemType: Union[int, str, np.uint], idx: np.ndarray) -> np.ndarray:
        """ Return the meshIO node ordering for a given element type
        """

        if isinstance(elemType, (int, np.integer)):
            elemType = self._gmsh_typing[int(elemType)]

        # 0D/1D/2D elements
        if elemType.startswith(('vertex', 'line', 'triangle', 'quad')):
            return idx

        # Check if we have a fixed ordering
        if elemType in self._meshio_ordering:
            return idx[:, self._meshio_ordering[elemType]]

        # Check if we are requesting higher-order simplices than currently implemented
        if not elemType.startswith('hexahedron'):
            raise ValueError(f'Unknown element type {elemType}')

        # For hexahedrons with analytic ordering
        nNodes = 8 if elemType.partition('hexahedron')[2] == '' else int(elemType.partition('hexahedron')[2])

        if self.deviation(nNodes ** (1/3) - 1) < self.deviation((nNodes-8)/12 + 1):
            nGeo = round(nNodes ** (1/3) - 1)
            incomplete = False
        else:
            nGeo = round((nNodes-8)/12 + 1)
            incomplete = True

        ordering = HEXREORDER(nGeo, incomplete=incomplete)
        return idx[:, ordering]

    def ordering_meshio_to_gmsh(self, elemType: Union[int, str, np.uint], idx: np.ndarray) -> np.ndarray:
        """ Return the Gmsh node ordering for a given element type
            > Inverse of ordering_gmsh_to_meshio
        """

        if isinstance(elemType, (int, np.integer)):
            elemType = self._gmsh_typing[int(elemType)]

        # 0D/1D/2D elements
        if elemType.startswith(('vertex', 'line', 'triangle', 'quad')):
            return idx

        # Check if we have a fixed ordering
        if elemType in self._meshio_ordering:
            perm      = np.asarray(self._meshio_ordering[elemType], dtype=int)  # meshio <- gmsh
            inv       = np.empty_like(perm)
            inv[perm] = np.arange(perm.size, dtype=int)                    # gmsh <- meshio
            return idx[:, inv]

        # Check if we are requesting higher-order simplices than currently implemented
        if not elemType.startswith('hexahedron'):
            raise ValueError(f'Unknown element type {elemType}')

        # For hexahedrons with analytic ordering
        nNodes = 8 if elemType.partition('hexahedron')[2] == '' else int(elemType.partition('hexahedron')[2])

        if self.deviation(nNodes ** (1/3) - 1) < self.deviation((nNodes - 8)/12 + 1):
            nGeo = round(nNodes ** (1/3) - 1)
            incomplete = False
        else:
            nGeo = round((nNodes - 8)/12 + 1)
            incomplete = True

        perm      = np.asarray(HEXREORDER(nGeo, incomplete=incomplete), dtype=int)
        inv       = np.empty_like(perm)
        inv[perm] = np.arange(perm.size, dtype=int)
        return idx[:, inv]

    def deviation(self, x: float) -> float:
        return abs(x - round(x))

    # INFO: Alternative implementation
    # def _compute_hexahedron_meshio_order(self, p: int, recursive: Optional[bool] = False) -> List[int]:
    #     # 1) Corner nodes
    #     mapping = list(range(8))
    #     if p == 1:
    #         return mapping
    #
    #     # Permutation for the 12 edge blocks in meshio ordering.
    #     GmshToMeshioEdgePerm = [0, 3, 5, 1, 8, 10, 11, 9, 2, 4, 6, 7]
    #     # Permutation for the 6 face blocks.
    #     GmshToMeshioFacePerm = [2, 3, 1, 4, 0, 5]
    #
    #     # 2) Edge nodes
    #     nNodeEdgeEdge   = p - 1
    #     nNodeEdgeTotal  = 12 * nNodeEdgeEdge
    #     gmshEdgeNodes   = list(range(8, 8 + nNodeEdgeTotal))
    #     # Partition edge nodes into 12 blocks.
    #     blockEdge       = [gmshEdgeNodes[i * nNodeEdgeEdge : (i + 1) * nNodeEdgeEdge] for i in range(12)]
    #     # Permute edge blocks to align with meshio order
    #     blockEdgeOrient = [blockEdge[i] for i in GmshToMeshioEdgePerm]
    #     blockEdgeOrient = [node for block in blockEdgeOrient for node in block]
    #     mapping.extend(blockEdgeOrient)
    #
    #     # 3) Face nodes
    #     nNodeFaceFace   = (p - 1) ** 2
    #     nNodeFaceTotal  = 6 * nNodeFaceFace
    #     startFaceNode   = 8 + nNodeEdgeTotal
    #     gmshFaceNodes   = list(range(startFaceNode, startFaceNode + nNodeFaceTotal))
    #     # Partition face nodes into 6 blocks.
    #     blockFace       = [gmshFaceNodes[i * nNodeFaceFace : (i + 1) * nNodeFaceFace] for i in range(6)]
    #     # Permuted face blocks to align with meshio order
    #     blockFaceOrient = [blockFace[i] for i in GmshToMeshioFacePerm]
    #     blockFaceOrient = [node for block in blockFaceOrient for node in block]
    #     mapping.extend(blockFaceOrient)
    #
    #     # 4) Interior nodes.
    #     # -- Interior nodes (recursive approach for p >= 4) --
    #     nNodeInterior = (p - 1) ** 3
    #     start_interior = startFaceNode + nNodeFaceTotal
    #
    #     if p <= 3:
    #         # For p <= 3, just append in natural order
    #         interior_nodes = list(range(start_interior, start_interior + nNodeInterior))
    #         mapping.extend(interior_nodes)
    #     # If we are the outermost call, we need to handle the recursive case
    #     elif not recursive:
    #         # General chunk-based approach for p >= 4
    #         remainder      = p - 2
    #         subcubeIndices = []
    #         currentOffset  = start_interior
    #
    #         # Repeatedly carve out sub-cubes (of order=3) until remainder used up
    #         while remainder >= 2:
    #             currenMap = self._compute_hexahedron_meshio_order(remainder, recursive=True)
    #             offsetMap = [currentOffset + node for node in currenMap]
    #             subcubeIndices.extend(offsetMap)
    #             currentOffset += len(currenMap)
    #             remainder     -= 2
    #
    #         mapping.extend(subcubeIndices)
    #
    #     return mapping

    def typing_gambit_to_meshio(self, elemType: Union[int, str, np.uint]) -> str:
        """
        Return the meshIO element type for a given Gambit element type
        """
        if isinstance(elemType, (int, np.integer)):
            return self._gambit_typing[int(elemType)]

        raise ValueError(f'Unknown element type {elemType}')

    def ordering_gambit_to_meshio(self, elemType: Union[int, str, np.uint], idx: np.ndarray) -> np.ndarray:
        """
        Return the meshIO node ordering for a given element type
        """
        if isinstance(elemType, (int, np.integer)):
            elemType = self._gambit_typing[int(elemType)]

        # 0D/1D/2D elements
        if elemType.startswith(('vertex', 'line', 'triangle', 'quad')):
            return idx

        # Check if we have a fixed ordering
        if elemType in self._gambit_ordering:
            return idx[self._gambit_ordering[elemType]]

        raise ValueError(f'Unknown element type {elemType}')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import meshio
import numpy as np
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# Monkey-patching MeshIO
# ==================================================================================================================================


def convertSerendipityToFullLagrange(mesh: meshio.Mesh) -> meshio.Mesh:
    """ Some GMSH files contain incomplete elements, fix them here
    """
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    from pyhope.common.common_tools import allocate_or_resize
    from pyhope.mesh.elements.elements_shapefunctions import ShapeFunctions
    from pyhope.mesh.elements.elements_ordering import ElementInfo
    # ------------------------------------------------------

    # Check the mesh contains second-order incomplete elements
    serendipityElems = ['quad8', 'hexahedron20', 'wedge15', 'pyramid13']
    if not any(s for s in mesh.cells_dict.keys() if s in serendipityElems):
        return mesh

    hopout.routine('Converting serendipity to full langrange mesh')

    # Copy original points
    points    = mesh.points.copy()

    # Prepare new cell blocks
    elems_new = {}
    shapefunctions = ShapeFunctions()
    elementinfo    = ElementInfo()

    for cell in mesh.cells:
        ctype, cdata = cell.type, cell.data

        # Valid cell type
        if ctype not in serendipityElems:
            elems_new[ctype] = cdata
            continue

        match ctype:
            # Remove surface elements
            case 'quad8':
                continue

            case 'hexahedron20':

                # Get number of hexahedrons which have to be converted
                nHex20    = len(cdata)

                faces     = ('x-', 'x+', 'y-', 'y+', 'z-', 'z+')
                nFaces    = len(faces)

                N         = [np.array(()) for _ in range(nFaces + 1)]
                faceNodes = [list()       for _ in faces]  # noqa: E272

                # preallocate the arrays for the new points and elements
                nPoints_old = len(points)
                nNewPoints  = (nFaces + 1) * nHex20
                points      = np.resize(points, (nPoints_old + nNewPoints, 3))

                # Allocate arrays if they do not exist. Else, resize them
                (elems_new, hex27_start) = allocate_or_resize(elems_new, 'hexahedron27', (nHex20,   27))
                (elems_new, quad9_start) = allocate_or_resize(elems_new, 'quad9',        (nHex20*6,  9))

                for iFace, face in enumerate(faces):
                    # Face parameters are the same as for the 27-node hexahedron
                    xi, eta, zeta    = elementinfo.faces_to_params('hexahedron20')[face]
                    faceNodes[iFace] = elementinfo.faces_to_nodes( 'hexahedron20')[face]

                    # Evaluate the quadratic shape function at the face center
                    N[iFace] = shapefunctions.evaluate(ctype, xi, eta, zeta)
                # Append the center node
                N[-1] = shapefunctions.evaluate(ctype, 0, 0, 0)

                # Loop over all hexahedrons
                for iElem, elem in enumerate(cdata):
                    # Create the 6 face mid-points
                    for iFace, face in enumerate(faces):
                        center = np.dot(N[iFace], mesh.points[elem])
                        points[nPoints_old + iFace, :]  = center

                        # Take the existing 8 face nodes and append the new center node
                        subFace = elem[faceNodes[iFace][:8]].tolist()
                        subFace.append(nPoints_old + iFace)
                        elems_new['quad9'][quad9_start + iElem * nFaces + iFace] = np.array(subFace, dtype=np.uint)

                    # Evaluate the quadratic shape function at the volume center
                    center = np.dot(N[-1], mesh.points[elem])
                    points[nPoints_old + nFaces, :] = center

                    # Create the volume element
                    subElem = elem.tolist()

                    # Append the 6 face center and the volume center
                    subElem.extend(range(nPoints_old, nPoints_old + nFaces + 1))
                    elems_new['hexahedron27'][hex27_start + iElem] = np.array(subElem, dtype=np.uint)

                    # Increment counter with number of added points
                    nPoints_old += nFaces + 1

            case 'wedge15':

                # Get number of hexahedrons which have to be converted
                nWed15    = len(cdata)

                faces     = ('y-', 'x+', 'x-')  # square faces of element
                # faces     = ['x+']
                nFaces    = len(faces)

                N         = [np.array(()) for _ in range(nFaces + 1)]
                faceNodes = [list()       for _ in faces]  # noqa: E272

                # preallocate the arrays for the new points and elements
                nPoints_old = len(points)
                nNewPoints  = nFaces * nWed15
                points      = np.resize(points, (nPoints_old + nNewPoints, 3))

                # Allocate arrays if they do not exist. Else, resize them
                (elems_new, wed18_start) = allocate_or_resize(elems_new, 'wedge18',   (nWed15,   18))
                (elems_new, quad9_start) = allocate_or_resize(elems_new, 'quad9',     (nWed15*3,  9))

                for iFace, face in enumerate(faces):
                    if 'z' in face:
                        continue

                    # Face parameters are the same as for the 9-node quad
                    xi, eta, zeta    = (0., 0., 0.)
                    faceNodes[iFace] = elementinfo.faces_to_nodes('wedge15')[face]

                    # Evaluate the quadratic shape function at the face center
                    N[iFace] = shapefunctions.evaluate('quad8', xi, eta, zeta)

                # Loop over all hexahedrons
                for iElem, elem in enumerate(cdata):
                    # Create the 3 face mid-points
                    for iFace, face in enumerate(faces):
                        # 2nd order triangular faces are already present in the mesh
                        # Only for sanity as faces are already excluded from "faces" array
                        if 'z' in face:
                            continue
                        # 2nd order quadrilaterial faces
                        elif len(faceNodes[iFace]) == 8:
                            # Here, we are on the quads and not the actual element
                            center = np.dot(N[iFace], mesh.points[elem[faceNodes[iFace]]])
                            points[nPoints_old + iFace, :]  = center

                            # Take the existing 8 face nodes and append the new center node
                            subFace = elem[faceNodes[iFace][:8]].tolist()
                            subFace.append(nPoints_old + iFace)
                            elems_new['quad9'][quad9_start + iElem * nFaces + iFace] = np.array(subFace, dtype=np.uint)

                    # Create the volume element
                    subElem = elem.tolist()

                    # Append the 3 face center
                    subElem.extend(range(nPoints_old, nPoints_old + nFaces))
                    elems_new['wedge18'][wed18_start + iElem] = np.array(subElem, dtype=np.uint)

                    # Increment counter with number of added points
                    nPoints_old += nFaces

            case 'pyramid13':
                # Get number of hexahedrons which have to be converted
                nPyra13   = len(cdata)

                face      = 'z-'  # square faces of element
                faceNodes = elementinfo.faces_to_nodes('pyramid13')[face]

                # Preallocate the arrays for the new points and elements
                nPoints_old = len(points)
                nNewPoints = nPyra13  # Since there's only one face per element
                points = np.resize(points, (nPoints_old + nNewPoints, 3))

                # Allocate arrays if they do not exist. Else, resize them
                (elems_new, wed18_start) = allocate_or_resize(elems_new, 'pyramid14', (nPyra13, 14))
                (elems_new, quad9_start) = allocate_or_resize(elems_new, 'quad9',     (nPyra13,  9))

                # Face parameters for the 9-node quad
                xi, eta, zeta = (0., 0., 0.)
                N = shapefunctions.evaluate('quad8', xi, eta, zeta)

                # Loop over all hexahedrons
                for iElem, elem in enumerate(cdata):
                    # Create the face mid-point
                    if len(faceNodes) == 8:
                        # Here, we are on the quads and not the actual element
                        center = np.dot(N, mesh.points[elem[faceNodes]])
                        points[nPoints_old, :]  = center

                        # Take the existing 8 face nodes and append the new center node
                        subFace = elem[faceNodes[:8]].tolist()
                        subFace.append(nPoints_old)
                        elems_new['quad9'][quad9_start + iElem] = np.array(subFace, dtype=np.uint)

                    # Create the volume element
                    subElem = elem.tolist()

                    # Append the 3 face center
                    subElem.append(nPoints_old)
                    elems_new['pyramid14'][wed18_start + iElem] = np.array(subElem, dtype=np.uint)

                    # Increment counter with number of added points
                    nPoints_old += 1

    # At this point, the mesh does not contain boundary conditions / cell sets
    # > We add them in the BCCGNS function
    mesh   = meshio.Mesh(points = points,     # noqa: E251
                         cells  = elems_new)  # noqa: E251

    # hopout.sep()
    # hopout.info('CONVERTING SERENDIPITY TO FULL LANGRANGE MESH DONE!')
    # hopout.sep()

    return mesh



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def DefineMesh() -> None:
    """ Define general options for mesh generation / readin
    """
    # Local imports ----------------------------------------
    from pyhope.readintools.readintools import CreateInt, CreateIntArray, CreateRealArray, CreateSection, CreateStr
    from pyhope.readintools.readintools import CreateLogical, CreateReal
    from pyhope.readintools.readintools import CreateIntFromString, CreateIntOption
    from pyhope.mesh.mesh_vars import ELEMTYPE, MeshMode, MeshSort
    # ------------------------------------------------------

    CreateSection('Mesh')
    CreateIntFromString('Mode',                            help=f'Mesh generation mode [{", ".join(s.name for s in MeshMode)}]')
    CreateIntOption(    'Mode', number=MeshMode.Internal.value,  name=MeshMode.Internal.name)
    CreateIntOption(    'Mode', number=MeshMode.External.value,  name=MeshMode.External.name)
    # Internal mesh generator
    CreateInt(      'nZones',                              help='Number of mesh zones')
    CreateRealArray('Corner',         24,   multiple=True, help='Corner node positions: (/ x_1,y_1,z_1,, x_2,y_2,z_2,, ' +
                                                                                         '... ,, x_8,y_8,z_8/)')
    CreateRealArray('X0',              3,   multiple=True, help='Origin of a zone. Equivalent to a corner node.')
    CreateRealArray('DX',              3,   multiple=True, help='Extension of the zone in each spatial direction ' +
                                                                 'starting from the origin X0 corner node')
    CreateIntArray( 'nElems',          3,   multiple=True, help='Number of elements in each direction')
    CreateIntFromString('ElemType'      ,   multiple=True, default='hexahedron', help='Element type')
    for key, val in ELEMTYPE.name.items():
        # Only consider uncurved element types
        if val > 200:
            continue
        CreateIntOption('ElemType', number=val, name=key)
    # Gmsh
    CreateLogical(  'EliminateNearDuplicates', default=True, help='Enables elimination of near duplicate points')
    # External mesh readin through GMSH
    CreateStr(      'Filename',             multiple=True, help='Name of external mesh file')
    CreateLogical(  'MeshIsAlreadyCurved',  default=False, help='Enables mesh agglomeration')
    # Common settings
    CreateInt(      'NGeo'         ,        default=1,     help='Order of spline-reconstruction for curved surfaces')
    CreateInt(      'BoundaryOrder',        default=2,     help='Order of spline-reconstruction for curved surfaces (legacy)')
    # Periodicity
    CreateRealArray('vv',              3,   multiple=True, help='Vector for periodic BC')
    CreateLogical(  'doPeriodicCorrect',    default=False, help='Enables periodic correction')
    # Connections
    CreateIntFromString('MeshSorting',      default=MeshSort.SFC.name,
                                            help=f'Mesh sorting mode [{", ".join([s.name for s in MeshSort if s.value != 0] + ["None"])}]')  # noqa: E501
    CreateIntOption(    'MeshSorting', number=MeshSort.NONE.value , name=MeshSort.NONE.name)
    CreateIntOption(    'MeshSorting', number=MeshSort.SFC.value  , name=MeshSort.SFC.name)
    CreateIntOption(    'MeshSorting', number=MeshSort.IJK.value  , name=MeshSort.IJK.name)
    CreateIntOption(    'MeshSorting', number=MeshSort.LEX.value  , name=MeshSort.LEX.name)
    CreateIntOption(    'MeshSorting', number=MeshSort.Snake.value, name=MeshSort.Snake.name)
    CreateLogical(  'doSortIJK',            default=False, help='Sort the mesh elements along the I,J,K directions (legacy)')
    CreateLogical(  'doSplitToHex',         default=False, help='Split simplex elements into hexahedral elements')
    # Mortars
    CreateLogical(  'doMortars',            default=True,  help='Enables mortars')
    # Boundaries
    CreateSection('Boundaries')
    CreateStr(      'BoundaryName',         multiple=True, help='Name of domain boundary')
    CreateIntArray( 'BoundaryType',    4,   multiple=True, help='(/ Type, curveIndex, State, alpha /)')
    CreateIntArray( 'BCIndex',         6,   multiple=True, help='Index of BC for each boundary face')
    # Checking
    CreateSection('Mesh Checks')
    CreateLogical(  'CheckElemJacobians',   default=True,  help='Check the Jacobian and scaled Jacobian for each element')
    CreateLogical(  'CheckConnectivity'  ,  default=True,  help='Check if the side connectivity, including correct flip')
    CreateLogical(  'CheckWatertightness',  default=True,  help='Check if the mesh is watertight')
    CreateLogical(  'CheckSurfaceNormals',  default=True,  help='Check if the surface normals point outwards')
    # Transformation
    CreateSection('Transformation')
    CreateReal(      'meshScale',           default=1.0,                              help='Scale the mesh')
    CreateRealArray( 'meshTrans', nReals=3, default='(/0.,0.,0./)',                   help='Translate the mesh')
    CreateRealArray( 'meshRot',   nReals=9, default='(/1.,0.,0.,0.,1.,0.,0.,0.,1./)', help='Rotate the mesh around rotation center')
    CreateRealArray( 'meshRot3D',   nReals=3, default='(/0.,0.,0./)'                , help='Rotate the mesh around rotation center and coordiante axis, defined angle in degrees')
    CreateRealArray( 'meshRotCenter', nReals=3, default='(/0.,0.,0./)',               help='Rotate the mesh around rotation center')
    CreateStr(       'MeshPostDeform',   default='none',                              help='Mesh post-transformation template')
    # Stretching
    CreateSection('Stretching')
    CreateIntArray( 'StretchType',      3,   default='(/0,0,0/)', multiple=True,      help='Stretching type for individual '
                                                                                             'zone per spatial direction.')
    CreateRealArray( 'Factor',          3,   multiple=True, help='Stretching factor of zone for geometric stretching for '
                                                                                                 'each spatial direction.')
    CreateRealArray( 'l0',              3,   multiple=True, help='Smallest desired element in zone per spatial direction.')
    CreateRealArray( 'DXmaxToDXmin',    3,   multiple=True, help='Ratio between the smallest and largest element per spatial '
                                                                                                               'direction')
    # Edge connectivity
    CreateSection('Finite Element Method (FEM) Connectivity')
    CreateLogical(  'doFEMConnect',         default=False, help='Generate finite element method (FEM) connectivity')


def InitMesh() -> None:
    """ Readin general option for mesh generation / readin
    """
    # Local imports ----------------------------------------
    import pyhope.io.io_vars as io_vars
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.readintools.readintools import GetInt, GetIntFromStr, CountOption
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('INIT MESH...')

    mesh_vars.mode = GetIntFromStr('Mode')

    NGeo     = GetInt('NGeo')          if CountOption('NGeo')          else None  # noqa: E272
    BCOrder  = GetInt('BoundaryOrder') if CountOption('BoundaryOrder') else None  # noqa: E272

    if not NGeo and not BCOrder:
        mesh_vars.nGeo = 1
    elif NGeo and BCOrder and NGeo != BCOrder - 1:
        hopout.error('NGeo / BoundaryOrder must be equal to NGeo + 1!')
    else:
        if NGeo is not None:
            mesh_vars.nGeo = NGeo
        elif BCOrder is not None:
            mesh_vars.nGeo = BCOrder - 1

        if mesh_vars.nGeo < 1:
            hopout.error('Effective boundary order < 1. Try increasing the NGeo / BoundaryOrder parameter!')

    # Check if the requested output format can supported the requested polynomial order
    match io_vars.outputformat:
        case io_vars.MeshFormat.VTK.value:
            if mesh_vars.nGeo > 2:
                hopout.error('Output format VTK does not support polynomial order > 2!')

    # hopout.info('INIT MESH DONE!')


def GenerateMesh() -> None:
    """ Generate the mesh
        Mode 1 - Use internal mesh generator
        Mode 2 - Readin external mesh through GMSH
    """
    # Local imports ----------------------------------------
    import pyhope.mesh.mesh_vars as mesh_vars
    import pyhope.output.output as hopout
    from pyhope.mesh.mesh_builtin import MeshCartesian
    from pyhope.mesh.mesh_external import MeshExternal
    from pyhope.mesh.mesh_vars import MeshMode
    from pyhope.mesh.topology.mesh_splittohex import MeshSplitToHex
    from pyhope.mesh.topology.mesh_topology import MeshChangeElemType
    # ------------------------------------------------------

    hopout.separator()
    hopout.info('GENERATE MESH...')

    match mesh_vars.mode:
        case MeshMode.Internal.value:  # Internal Cartesian Mesh
            mesh = MeshCartesian()
        case MeshMode.External.value:  # External mesh
            mesh = MeshExternal()
        case _:  # Default
            hopout.error('Unknown mesh mode {}, exiting...'.format(mesh_vars.mode), traceback=True)

    # Split hexahedral elements if requested
    mesh = MeshChangeElemType(mesh)
    # Split simplex elements if requested
    mesh = MeshSplitToHex(mesh)
    mesh_vars.mesh = mesh

    # Final count
    nElems = 0
    for cellType in mesh.cells:
        if any(s in cellType.type for s in mesh_vars.ELEMTYPE.type.keys()):
            nElems += mesh.get_cells_type(cellType.type).shape[0]

    hopout.routine('Generated mesh with {} cells'.format(nElems))
    # hopout.sep()
    # hopout.info('GENERATE MESH DONE!')
    hopout.separator()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import os
import shutil
import platform
import subprocess
from importlib import metadata
from packaging.version import Version
from typing import Optional, cast
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def PkgsMetaData(pkgs, classifier) -> Optional[bool]:
    """ Check if the package contains a given classifier
    """
    try:
        meta = metadata.metadata(pkgs)
        classifiers = meta.get_all('Classifier', [])
        return classifier in classifiers

    except metadata.PackageNotFoundError:
        return None


def PkgsMetaVersion(pkgs) -> Optional[str]:
    """ Check the package version
    """
    try:
        version = metadata.version(pkgs)
        return version

    except metadata.PackageNotFoundError:
        return None


def PkgsCheckGmsh() -> None:
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    from pyhope.common.common import IsInteractive
    from pyhope.common.common_vars import Gitlab
    # ------------------------------------------------------

    # Check the current platform
    system = platform.system().lower()
    arch   = platform.machine().lower()

    gmsh_version  = PkgsMetaVersion('gmsh')
    if gmsh_version is None:
        # Gmsh is not installed
        if IsInteractive():
            if system in Gitlab.LIB_SUPPORT and arch in Gitlab.LIB_SUPPORT[system]:
                warning = 'Gmsh is not installed. For compatibility, the NRG Gmsh version will be installed. Continue? (Y/n):'
                response = input('\n' + hopout.warn(warning) + '\n')
                if response.lower() in ['yes', 'y', '']:
                    PkgsInstallGmsh(system, arch, version='nrg')
                    return None
            else:
                warning = 'Gmsh is not installed. As NRG does not provide a compatible Gmsh version,' + \
                          'the PyPI Gmsh version will be installed. Continue? (Y/n):'
                response = input('\n' + hopout.warn(warning) + '\n')
                if response.lower() in ['yes', 'y', '']:
                    PkgsInstallGmsh(system, arch, version='pypi')
                    return None
        else:
            hopout.error('Gmsh is not installed, exiting...')

    # Assume that newer versions have updated CGNS
    gmsh_version  = cast(str, gmsh_version)
    gmsh_expected = '5.0'
    if Version(gmsh_version) > Version(gmsh_expected):
        return None

    # Check if the installed version is the NRG version
    if PkgsMetaData('gmsh', 'Intended Audience :: NRG'):
        return None

    if system not in Gitlab.LIB_SUPPORT or arch not in Gitlab.LIB_SUPPORT[system]:
        warning = hopout.warn(f'Detected non-NRG Gmsh version on unsupported platform [{system}/{arch}]. ' +
                              'Functionality may be limited.')
        print(warning)
        return None

    if not PkgsMetaData('gmsh', 'Intended Audience :: NRG'):
        if IsInteractive():
            warning  = 'Detected Gmsh package uses an outdated CGNS (v3.4). For compatibility, ' + \
                       'the package will be uninstalled and replaced with the updated NRG GMSH ' + \
                       'version. Continue? (Y/n):'
            response = input('\n' + hopout.warn(warning) + '\n')
            if response.lower() in ['yes', 'y', '']:
                PkgsInstallGmsh(system, arch, version='nrg')
                return None
        else:
            warning = hopout.warn('Detected Gmsh package uses an outdated CGNS (v3.4). Functionality may be limited.')
            print(warning)
            return None


def PkgsInstallGmsh(system: str, arch: str, version: str) -> None:
    # Standard libraries -----------------------------------
    import sys
    import hashlib
    import tempfile
    # Local imports ----------------------------------------
    import pyhope.output.output as hopout
    from pyhope.common.common_vars import Gitlab
    # ------------------------------------------------------
    # Get our package manager
    # > Check if 'uv' is available
    command = None
    if shutil.which('uv') is not None:
        command = ['uv', 'pip']

    # > Check if 'python -m pip' is available
    try:
        subprocess.run([sys.executable, '-m', 'pip', '--version'], check=True,
                       stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        command = [sys.executable, '-m', 'pip']
    except subprocess.CalledProcessError:
        pass

    if command is None:
        hopout.warning('No package manager found, you are on your own...')
        return None

    if version == 'nrg':
        # Gitlab "python-gmsh" access
        lfs = 'yes'
        lib = 'gmsh-{}-py3-none-{}_{}.whl'.format(Gitlab.LIB_VERSION[system][arch], system, arch)

        # Create a temporary directory
        with tempfile.TemporaryDirectory() as path:
            # On macOS add major version string to filename an rename darwin to macosx in whl filename
            if system == 'darwin':
                mac_ver = platform.mac_ver()[0].split('.')[0]
                lib  = lib.replace('darwin', 'macosx')
                pkgs = os.path.join(path, lib.replace('macosx_', f'macosx_{mac_ver}_0_'))
            else:
                pkgs = os.path.join(path, lib)

            curl = [f'curl https://{Gitlab.LIB_GITLAB}/api/v4/projects/{Gitlab.LIB_PROJECT}/repository/files/{lib}/raw?lfs={lfs} --output {pkgs}']  # noqa: E501
            _ = subprocess.run(curl, check=True, shell=True)

            # Compare the hash
            # > Initialize a new sha256 hash
            sha256 = hashlib.sha256()
            with open(pkgs, 'rb') as f:
                # Read and update hash string value in blocks of 4K
                for chunk in iter(lambda: f.read(4096), b""):
                    sha256.update(chunk)

            if sha256.hexdigest() == Gitlab.LIB_SUPPORT[system][arch]:
                hopout.info('Hash matches, installing Gmsh wheel...')
            else:
                hopout.error('Hash mismatch, exiting...')

            # Remove the old version
            try:
                meta = metadata.metadata('gmsh')
                if meta is not None:
                    _ = subprocess.run(command + ['uninstall'] + ['gmsh'], check=True)  # noqa: E501

            except metadata.PackageNotFoundError:
                pass

            # Install the package in the current environment
            _ = subprocess.run(command + ['install'] + [pkgs], check=True)
    else:
        # Install the package in the current environment
        _ = subprocess.run(command + ['install'] + ['gmsh'], check=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from functools import cache
from typing import Final, Union
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
import scipy as sp
# from threadpoolctl import ThreadpoolController
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
from pyhope.common.common_numba import NUMBA_AVAILABLE
from pyhope.common.common_numba import jit, types
from pyhope.mesh.mesh_common import NDOFS_ELEM
# ==================================================================================================================================


def legendre_gauss_lobatto_nodes(order: int) -> tuple[np.ndarray, np.ndarray]:
    """ Return Legendre-Gauss-Lobatto nodes and weights for a given order in 1D
    """
    order -= 1
    # Special cases for small N
    if order == 1:
        return np.array((-1, 1)), np.array((1, 1))

    # Compute the initial guess for the LGL nodes (roots of P'_N)
    nodes = np.cos(np.pi * np.arange(order+1) / order)

    # Initialize the Legendre polynomial and its derivative
    p = np.zeros((order+1, order+1))

    # Iteratively solve for the LGL nodes using Newton's method
    xOld = 2 * np.ones_like(nodes)
    tol = 1e-14
    while np.max(np.abs(nodes - xOld)) > tol:
        xOld = nodes.copy()
        p[:, 0] = 1
        p[:, 1] = nodes
        for k in range(2, order+1):
            p[:, k] = ((2*k-1) * nodes * p[:, k-1] - (k-1) * p[:, k-2]) / k
        nodes -= (nodes * p[:, order] - p[:, order-1]) / (order * (p[:, order]))

    # The LGL nodes
    nodes = np.sort(nodes)

    # Compute the LGL weights
    weights = 2 / (order * (order + 1) * (p[:, order]**2))

    return nodes, weights


def legendre_gauss_nodes(order: int) -> tuple[np.ndarray, np.ndarray]:
    """ Return Legendre-Gauss nodes and weights for a given order in 1D
    """
    nodes, weights = np.polynomial.legendre.leggauss(order)
    return nodes, weights


@cache
def equi_nodes_prism(order: int) -> np.ndarray:
    """ Return equidistant nodes on a wedge/prism
    """
    xEq = np.linspace(-1, 1, num=order, dtype=np.float64)
    iZETA, iETA, iXI = np.indices((order, order, order))
    mask = iXI < (order - iETA)
    return np.vstack((xEq[iXI[mask]], xEq[iETA[mask]], xEq[iZETA[mask]]))


@cache
def equi_nodes_pyram(order: int) -> np.ndarray:
    """ Return equidistant nodes on a pyramid
    """
    xEq = np.linspace(-1, 1, num=order, dtype=np.float64)
    iZETA, iETA, iXI = np.indices((order, order, order))
    mask = (iXI < (order - iZETA)) & (iETA < (order - iZETA))
    return np.vstack((xEq[iXI[mask]], xEq[iETA[mask]], xEq[iZETA[mask]]))


@cache
def equi_nodes_tetra(order: int) -> np.ndarray:
    """ Return equidistant nodes on a tetrahedron
    """
    xEq = np.linspace(-1, 1, num=order, dtype=np.float64)
    iZETA, iETA, iXI = np.indices((order, order, order))
    mask = (iXI < (order - iETA - iZETA)) & (iETA < (order - iZETA))
    return np.vstack((xEq[iXI[mask]], xEq[iETA[mask]], xEq[iZETA[mask]]))


def barycentric_weights(_: int, xGP: np.ndarray) -> np.ndarray:
    """ Compute the barycentric weights for a given node set
        > Algorithm 30, Kopriva
    """
    # Create a difference matrix (x_i - x_j) for all i, j
    diff_matrix = xGP[:, np.newaxis] - xGP[np.newaxis, :]

    # Set the diagonal to 1 to avoid division by zero (diagonal elements will not be used)
    np.fill_diagonal(diff_matrix, 1.0)

    # Compute the product of all differences for each row (excluding the diagonal)
    wBary = np.prod(diff_matrix, axis=1)

    # Take the reciprocal to get the final barycentric weights
    # wBary = 1.0 / wBary

    return 1.0 / wBary


def polynomial_derivative_matrix(order: int, xGP: np.ndarray) -> np.ndarray:
    """ Compute the polynomial derivative matrix for a given node set
        > Algorithm 37, Kopriva
    """
    wBary = barycentric_weights(order, xGP)
    D     = np.zeros((order, order), dtype=float)

    for iLagrange in range(order):
        for iGP in range(order):
            if iLagrange != iGP:
                D[iGP, iLagrange] = wBary[iLagrange]/(wBary[iGP]*(xGP[iGP]-xGP[iLagrange]))
                D[iGP, iGP      ] = D[iGP, iGP] - D[iGP, iLagrange]

    return D


def lagrange_interpolation_polys(x: Union[float, np.ndarray], order: int, xGP: np.ndarray, wBary: np.ndarray) -> np.ndarray:
    """ Computes all Lagrange functions evaluated at position x in [-1;1]
        > Algorithm 34, Kopriva
    """
    # Equal points need special treatment
    lagrange = np.zeros(order)
    for iGP in range(order):
        if abs(x - xGP[iGP]) < 1.E-14:
            lagrange[iGP] = 1
            return lagrange

    tmp = 0.
    for iGP in range(order):
        lagrange[iGP] = wBary[iGP] / (x-xGP[iGP])
        tmp += lagrange[iGP]

    # Normalize
    # lagrange = lagrange/tmp

    return lagrange/tmp


def calc_vandermonde(n_In: int, n_Out: int, wBary_In: np.ndarray, xi_In: np.ndarray, xi_Out: np.ndarray) -> np.ndarray:
    """ Build a 1D Vandermonde matrix using the Lagrange basis functions of degree N_In,
        evaluated at the interpolation points xi_Out
    """
    Vdm = np.zeros((n_Out, n_In))
    for iXI in range(n_Out):
        Vdm[iXI, :] = lagrange_interpolation_polys(xi_Out[iXI], n_In, xi_In, wBary_In)
    return Vdm


#  def compute_cols_prism(a, b, c, n):
#      cols = []
#      for iZETA in range(n):
#          fZETA = sp.special.jacobi(iZETA, 0, 0)(c)
#          for iETA in range(n):
#              fETA = sp.special.jacobi(iETA, 0, 0)(b)
#              for iXI in range(n - iETA):
#                  fXI = sp.special.jacobi(iXI, 2*iETA + 1, 0)(a)
#                  col = np.sqrt(2.) * fXI * fETA * fZETA * (1 - a) ** iETA
#                  cols.append(col)
#      return np.array(cols).T
#
#
#  def compute_cols_tetra(a, b, c, n):
#      cols = []
#      for iZETA in range(n):
#          for iETA in range(n - iZETA):
#              for iXI in range(n - iETA - iZETA):
#                  fXI   = sp.special.jacobi(iXI  , 0             , 0)(a)
#                  fETA  = sp.special.jacobi(iETA , 2*iETA + 1    , 0)(b)
#                  fZETA = sp.special.jacobi(iZETA, 2*(iETA+iXI)+2, 0)(c)
#                  col = 2.*np.sqrt(2.) * fXI * fETA * fZETA * (1 - b) ** iXI * (1 - c) ** (iXI+iETA)
#                  cols.append(col)
#      return np.array(cols).T
#
#
#  def calc_vandermonde_prism(n_In: int, n_Out: int, xi_In: np.ndarray, xi_Out: np.ndarray) -> np.ndarray:
#      """ Build a 3D Vandermonde matrix using the PKD basis function,
#          evaluated at the interpolation points xi_Out (build of Jacobi polynomials of degree N_In)
#      """
#      a_in  = xi_In[0, :]
#      b_in  = np.where((1 - xi_In[0, :]) <= 1.e-12, -1.0, 2 * (1 + xi_In[1, :]) / (1. - xi_In[0, :] + 1.e-20) - 1.)
#      c_in  = xi_In[2, :]
#
#      a_out = xi_Out[0, :]
#      b_out = np.where((1 - xi_Out[0, :]) <= 1.e-12, -1.0, 2 * (1 + xi_Out[1, :]) / (1. - xi_Out[0, :] + 1.e-20) - 1.)
#      c_out = xi_Out[2, :]
#
#      if n_In >= n_Out:
#          Vdm_tmp = compute_cols_prism(a_in , b_in , c_in , n_In)
#          Vdm     = compute_cols_prism(a_out, b_out, c_out, n_In)
#          Vdm     = np.linalg.inv(Vdm_tmp) @ Vdm
#      else:
#          Vdm_tmp = compute_cols_prism(a_out, b_out, c_out, n_Out)
#          Vdm     = compute_cols_prism(a_in , b_in , c_in , n_Out)
#          Vdm     = Vdm @ np.linalg.inv(Vdm_tmp)
#
#      return Vdm
#
#
#  def calc_vandermonde_tetra(n_In: int, n_Out: int, xi_In: np.ndarray, xi_Out: np.ndarray) -> np.ndarray:
#      """ Build a 3D Vandermonde matrix using the PKD basis function,
#          evaluated at the interpolation points xi_Out (build of Jacobi polynomials of degree N_In)
#      """
#      a_in  = np.where(np.abs(xi_In[1, :] + xi_In[2, :]) <= 1.e-12, -1.0, 2 * (1 + xi_In[0, :]) / (- xi_In[1, :] - xi_In[2, :] + 1.e-20) - 1.)  # noqa: E501
#      b_in  = np.where((1 - xi_In[2, :]) <= 1.e-12, -1.0, 2 * (1 + xi_In[1, :]) / (1. - xi_In[2, :] + 1.e-20) - 1.)
#      c_in  = xi_In[2, :]
#
#      a_out = np.where(np.abs(xi_Out[1, :] + xi_Out[2, :]) <= 1.e-12, -1.0, 2 * (1 + xi_Out[0, :]) / (- xi_Out[1, :] - xi_Out[2, :] + 1.e-20) - 1.)  # noqa: E501
#      b_out = np.where((1 - xi_Out[2, :]) <= 1.e-12, -1.0, 2 * (1 + xi_Out[1, :]) / (1. - xi_Out[2, :] + 1.e-20) - 1.)
#      c_out = xi_Out[2, :]
#
#      if n_In >= n_Out:
#          Vdm_tmp = compute_cols_tetra(a_in , b_in , c_in , n_In)
#          Vdm     = compute_cols_tetra(a_out, b_out, c_out, n_In)
#          Vdm     = np.linalg.inv(Vdm_tmp) @ Vdm
#      else:
#          Vdm_tmp = compute_cols_tetra(a_out, b_out, c_out, n_Out)
#          Vdm     = compute_cols_tetra(a_in , b_in , c_in , n_Out)
#          Vdm     = Vdm @ np.linalg.inv(Vdm_tmp)
#
#      return Vdm


def polynomial_derivative_matrix_prism(order: int, xGP: np.ndarray) -> np.ndarray:
    """ Compute the polynomial derivative matrix for a prism
    """
    a: Final[np.ndarray]  = xGP[0, :]
    b: Final[np.ndarray]  = np.where((1 - xGP[0, :]) <= 1.e-12, -1.0, 2 * (1 + xGP[1, :]) / (1. - xGP[0, :] + 1.e-20) - 1.)
    c: Final[np.ndarray]  = xGP[2, :]

    nDOFs = NDOFS_ELEM(106, order-1)
    Vdm = np.zeros((   nDOFs, nDOFs))
    Vr  = np.zeros((   nDOFs, nDOFs))
    Vs  = np.zeros((   nDOFs, nDOFs))
    Vt  = np.zeros((   nDOFs, nDOFs))
    D   = np.zeros((3, nDOFs, nDOFs))

    # Precompute required Jacobi polynomials and derivatives
    # fZETA(i) = P_i^(0,0)(c), dfZETA(i) = 0.5*(i+1)*P_{i-1}^{(1,1)}(c)
    fZETA_all  = np.array([          sp.special.eval_jacobi(i  , 0, 0, c) for i in range(order)])
    dfZETA_all = np.array([0.5*(i+1)*sp.special.eval_jacobi(i-1, 1, 1, c) for i in range(order)])
    # fETA(j) = P_j^(0,0)(b), dfETA(j) = 0.5*(j+1)*P_{j-1}^{(1,1)}(b)
    fETA_all   = np.array([          sp.special.eval_jacobi(j  , 0, 0, b) for j in range(order)])
    dfETA_all  = np.array([0.5*(j+1)*sp.special.eval_jacobi(j-1, 1, 1, b) for j in range(order)])

    jacobi_xi_polys  = [[sp.special.jacobi(i, 2*j + 1, 0) for i in range(order - j)] for j in range(order)]
    jacobi_xi_derivs = [[p.deriv() for p in row] for row in jacobi_xi_polys]

    # Precompute constants
    sqrt2    = np.sqrt(2.)
    base_pa  = 0.5*(1 - a)
    base_pb  = 0.5*(1 + b)

    pow1ma   = [np.ones_like(a)]  # (1 - a)**k
    for _ in range(1, order+1):
        pow1ma.append(pow1ma[-1] * (1 - a))

    iX = 0
    for iZETA in range(order):
        fZETA, dfZETA = fZETA_all[iZETA], dfZETA_all[iZETA]
        for iETA in range(order):
            fETA, dfETA = fETA_all[iETA], dfETA_all[iETA]
            # fXI(ix) = P_ix^(2*iETA+1, 0)(a)
            # dfXI(ix)= 0.5*(ix + 2*iETA + 2) * P_{ix-1}^{(2*iETA+2, 1)}(a), with dfXI(0)=0
            for iXI in range(order - iETA):
                fXI  = jacobi_xi_polys[ iETA][iXI](a)
                dfXI = jacobi_xi_derivs[iETA][iXI](a)

                # powfac = (1 - a) ** iETA
                powfac = pow1ma[iETA]

                # Fill Vdm columns
                Vdm[:, iX] = sqrt2 * fXI * fETA * fZETA * powfac

                # Vr term
                Vr[:, iX] = dfETA * fXI * fZETA
                if iETA > 0:
                    Vr[:, iX] *= (base_pa**(iETA-1))

                # Vs term
                Vs[:, iX] = dfETA * (fXI * base_pb) * fZETA
                if iETA > 0:
                    Vs[:, iX] *= (base_pa**(iETA-1))
                tmp = dfXI * (base_pa**iETA)
                if iETA > 0:
                    tmp -= 0.5*iETA*fXI*(base_pa**(iETA-1))
                Vs[:, iX] += fETA*tmp*fZETA

                # Vt term
                Vt[:, iX] = sqrt2*dfZETA*fETA*fXI*powfac

                # Final scaling
                scale = 2.**(iETA+0.5)
                Vr[:, iX] *= scale
                Vs[:, iX] *= scale

                iX += 1

    # PERF: Need to use threadpoolctl here as BLAS has issues with multiprocessing
    #       > Workaround by restricting to one thread during concurrent execution
    # with ThreadpoolController().limit(limits=1, user_api='blas'):
    #     sVdm = np.linalg.inv(Vdm)
    # D[1, :, :] = (Vr @ sVdm).T
    # D[0, :, :] = (Vs @ sVdm).T
    # D[2, :, :] = (Vt @ sVdm).T

    # PERF: Solve with LU factorization instead of explicitly inverting, disabling input matrix checks
    VT = Vdm.T
    lu, piv    = sp.linalg.lu_factor(VT,             check_finite=False)
    D[1, :, :] = sp.linalg.lu_solve((lu, piv), Vr.T, check_finite=False)
    D[0, :, :] = sp.linalg.lu_solve((lu, piv), Vs.T, check_finite=False)
    D[2, :, :] = sp.linalg.lu_solve((lu, piv), Vt.T, check_finite=False)

    return D


def polynomial_derivative_matrix_pyram(order: int, xGP: np.ndarray) -> np.ndarray:
    """ Compute the polynomial derivative matrix for a pyramid
    """
    a  = np.where((1 - xGP[2, :]) <= 1.e-12, -1.0, 2 * (1 + xGP[0, :]) / (1. - xGP[2, :] + 1.e-20) - 1.)
    b  = np.where((1 - xGP[2, :]) <= 1.e-12, -1.0, 2 * (1 + xGP[1, :]) / (1. - xGP[2, :] + 1.e-20) - 1.)
    c  = xGP[2, :]

    nDOFs = NDOFS_ELEM(105, order-1)
    Vdm = np.zeros((   nDOFs, nDOFs))
    Vr  = np.zeros((   nDOFs, nDOFs))
    Vs  = np.zeros((   nDOFs, nDOFs))
    Vt  = np.zeros((   nDOFs, nDOFs))
    D   = np.zeros((3, nDOFs, nDOFs))

    iX = 0
    for iZETA in range(order):
        for iETA in range(order - iZETA):
            for iXI in range(order - iZETA):
                # fXI  = P_{iXI }^{(0,0)}(a), dfXI  = 0.5*(iXI +1)*P_{iXI -1}^{(1,1)}(a)
                fXI    = sp.special.jacobi(iXI  , 0, 0)(a)
                dfXI   = sp.special.jacobi(iXI  , 0, 0).deriv()(a)

                # fETA = P_{iETA}^{(0,0)}(b), dfETA = 0.5*(iETA+1)*P_{iETA-1}^{(1,1)}(b)
                fETA   = sp.special.jacobi(iETA , 0, 0)(b)
                dfETA  = sp.special.jacobi(iETA , 0, 0).deriv()(b)

                # fZETA is the dependent direction: alpha = 2*(iXI+iETA)+2, beta = 0
                fZETA  = sp.special.jacobi(iZETA, 2*(iXI+iETA)+2, 0)(c)
                dfZETA = sp.special.jacobi(iZETA, 2*(iXI+iETA)+2, 0).deriv()(c)

                # Fill Vdm columns
                Vdm[:, iX] = 2 * fXI * fETA * fZETA * (1-c)**(iXI+iETA)

                # Vr term
                Vr[:, iX] = dfXI*fETA*fZETA
                if iXI+iETA > 0:
                    Vr[:, iX] *= 2*(1-c)**(iXI+iETA-1)

                # Vs term
                Vs[:, iX] = fXI*dfETA*fZETA
                if iXI+iETA > 0:
                    Vs[:, iX] *= 2*(1-c)**(iXI+iETA-1)

                # Vt term
                tmp = (1-c)**(iXI+iETA)*dfZETA
                if iXI+iETA > 0:
                    tmp -= (iETA+iXI)*(1-c)**(iETA+iXI-1)*fZETA
                Vt[:, iX] = 2*tmp*fXI*fETA
                Vt[:, iX] += (1+a)*Vr[:, iX] + (1+b)*Vs[:, iX]

                # Final scaling
                Vr[:, iX] *= 2
                Vs[:, iX] *= 2

                iX += 1

    # PERF: Need to use threadpoolctl here as BLAS has issues with multiprocessing
    #       > Workaround by restricting to one thread during concurrent execution
    # with ThreadpoolController().limit(limits=1, user_api='blas'):
    #     sVdm = np.linalg.inv(Vdm)
    # D[0, :, :] = (Vr @ sVdm).T
    # D[1, :, :] = (Vs @ sVdm).T
    # D[2, :, :] = (Vt @ sVdm).T

    # PERF: Solve with LU factorization instead of explicitly inverting, disabling input matrix checks
    VT = Vdm.T
    lu, piv    = sp.linalg.lu_factor(VT,             check_finite=False)
    D[0, :, :] = sp.linalg.lu_solve((lu, piv), Vr.T, check_finite=False)
    D[1, :, :] = sp.linalg.lu_solve((lu, piv), Vs.T, check_finite=False)
    D[2, :, :] = sp.linalg.lu_solve((lu, piv), Vt.T, check_finite=False)

    return D


def polynomial_derivative_matrix_tetra(order: int, xGP: np.ndarray) -> np.ndarray:
    """ Compute the polynomial derivative matrix for a tetra
    """
    a  = np.where(np.abs(xGP[1, :] + xGP[2, :]) <= 1.e-12, -1.0, 2 * (1 + xGP[0, :]) / (- xGP[1, :] - xGP[2, :] + 1.e-20) - 1.)
    b  = np.where((1 - xGP[2, :]) <= 1.e-12, -1.0, 2 * (1 + xGP[1, :]) / (1. - xGP[2, :] + 1.e-20) - 1.)
    c  = xGP[2, :]

    nDOFs = NDOFS_ELEM(104, order-1)
    Vdm = np.zeros((   nDOFs, nDOFs))
    Vr  = np.zeros((   nDOFs, nDOFs))
    Vs  = np.zeros((   nDOFs, nDOFs))
    Vt  = np.zeros((   nDOFs, nDOFs))
    D   = np.zeros((3, nDOFs, nDOFs))

    # Precompute constants
    sqrt2    = np.sqrt(2.)

    iX = 0
    for iZETA in range(order):
        for iETA in range(order - iZETA):
            for iXI in range(order - iETA - iZETA):
                # fXI(a), dfXI(a)
                fXI    = sp.special.jacobi(iXI  , 0             , 0)(a)
                dfXI   = sp.special.jacobi(iXI  , 0             , 0).deriv()(a)

                # fETA(b; alpha=2*iXI+1), dfETA(b)
                fETA   = sp.special.jacobi(iETA , 2*iXI+1       , 0)(b)
                dfETA  = sp.special.jacobi(iETA , 2*iXI+1       , 0).deriv()(b)

                # fZETA(c; alpha=2*(iXI+iETA)+2), dfZETA(c)
                fZETA  = sp.special.jacobi(iZETA, 2*(iXI+iETA)+2, 0)(c)
                dfZETA = sp.special.jacobi(iZETA, 2*(iXI+iETA)+2, 0).deriv()(c)

                # Fill Vdm columns
                Vdm[:, iX] = 2 * sqrt2 * fXI * fETA * fZETA * (1-b)**iXI * (1-c)**(iXI+iETA)

                # Vr term
                Vr[:, iX] = 2 * sqrt2 * dfXI * fETA * fZETA
                if iXI > 0:
                    Vr[:, iX] *= ((0.5*(1-b))**(iXI-1))
                if iXI+iETA > 0:
                    Vr[:, iX] *= ((0.5*(1-c))**(iXI+iETA-1))

                # Vs term
                tmp = (1-b)**iXI*dfETA
                if iXI > 0:
                    tmp -= iXI*(1-b)**(iXI-1)*fETA
                if iXI+iETA > 0:
                    tmp *= (1-c)**(iXI+iETA-1)
                tmp *= 2.*sqrt2*fXI*fZETA
                Vs[:, iX] = tmp + Vr[:, iX] * (1+a)

                # Vt term
                Vt[:, iX] = 2*(1+a)*Vr[:, iX] + (1+b)*tmp
                tmp = dfZETA*(1-c)**(iETA+iXI)
                if iXI+iETA > 0:
                    tmp -= (iXI+iETA)*(1-c)**(iETA+iXI-1)*fZETA
                tmp *= 2*sqrt2*fXI*fETA*(1-b)**iXI
                Vt[:, iX] += tmp

                Vr[:, iX] *= 4
                Vs[:, iX] *= 2
                iX += 1

    # PERF: Need to use threadpoolctl here as BLAS has issues with multiprocessing
    #       > Workaround by restricting to one thread during concurrent execution
    # with ThreadpoolController().limit(limits=1, user_api='blas'):
    #     sVdm = np.linalg.inv(Vdm)
    # D[0, :, :] = (Vr @ sVdm).T
    # D[1, :, :] = (Vs @ sVdm).T
    # D[2, :, :] = (Vt @ sVdm).T

    # PERF: Solve with LU factorization instead of explicitly inverting, disabling input matrix checks
    VT = Vdm.T
    lu, piv    = sp.linalg.lu_factor(VT,             check_finite=False)
    D[0, :, :] = sp.linalg.lu_solve((lu, piv), Vr.T, check_finite=False)
    D[1, :, :] = sp.linalg.lu_solve((lu, piv), Vs.T, check_finite=False)
    D[2, :, :] = sp.linalg.lu_solve((lu, piv), Vt.T, check_finite=False)

    return D


if not NUMBA_AVAILABLE:
    def change_basis_3D(Vdm: np.ndarray, x3D_In: np.ndarray) -> np.ndarray:
        """ Interpolate a 3D tensor product Lagrange basis defined by (N_in+1) 1D interpolation point positions xi_In(0:N_In)
            to another 3D tensor product node positions (number of nodes N_out+1)
            defined by (N_out+1) interpolation point  positions xi_Out(0:N_Out)
            xi is defined in the 1D reference element xi=[-1,1]
        """
        # INFO: numpy-version, not compatible with numba
        # First contraction along the iN_In axis (axis 1 of Vdm, axis 1 of x3D_In)
        # x3D_Buf1 = np.tensordot(Vdm, x3D_In , axes=(1, 1))
        # x3D_Buf1 = np.moveaxis(x3D_Buf1, 0, 1)  # Correct the shape to (dim1, n_Out, n_In, n_In)
        n_In     = Vdm.shape[1]
        X1       = x3D_In.reshape(x3D_In.shape[0], n_In, n_In * n_In)
        x3D_Buf1 = (Vdm @ X1).reshape(x3D_In.shape[0], Vdm.shape[0], n_In, n_In)  # (dim1, n_Out, n_In, n_In)

        # Second contraction along the jN_In axis (axis 1 of Vdm, axis 2 of x3D_Buf1)
        x3D_Buf2 = np.tensordot(Vdm, x3D_Buf1, axes=(1, 2))
        x3D_Buf2 = np.moveaxis(x3D_Buf2, 0, 2)  # Correct the shape to  (dim1, n_Out, n_Out, n_In)

        # Third contraction along the kN_In axis (axis 1 of Vdm, axis 3 of x3D_Buf2)
        # x3D_Out  = np.tensordot(Vdm, x3D_Buf2, axes=(1, 3))
        # x3D_Out  = np.moveaxis(x3D_Out , 0, 3)  # Correct the shape to (dim1, n_Out, n_Out, n_Out)
        # x3D_Out = x3D_Buf2 @ Vdm.T
        # PERF: This is actually slower than the individual contractions
        # x3D_Out  = np.einsum('pi,qj,rk,dijk->dpqr', Vdm, Vdm, Vdm, x3D_In, optimize=True)

        return x3D_Buf2 @ Vdm.T

else:
    @jit((types.float64[:, :, :, ::1])(types.float64[:, ::1], types.float64[:, :, :, :]), nopython=True, cache=True, nogil=True)
    def change_basis_3D(Vdm: np.ndarray, x3D_In: np.ndarray) -> np.ndarray:
        """ Interpolate a 3D tensor product Lagrange basis defined by (N_in+1) 1D interpolation point positions xi_In(0:N_In)
            to another 3D tensor product node positions (number of nodes N_out+1)
            defined by (N_out+1) interpolation point  positions xi_Out(0:N_Out)
            xi is defined in the 1D reference element xi=[-1,1]
        """
        dim1, n_In, _, _ = x3D_In.shape
        n_Out = Vdm.shape[0]

        # First contraction along the iN_In axis (axis 1 of Vdm, axis 1 of x3D_In)
        buf1 = np.zeros((dim1, n_Out, n_In, n_In), dtype=x3D_In.dtype)
        for d in range(dim1):
            for k in range(n_In):
                buf1[d, :, :, k] = Vdm @ np.ascontiguousarray(x3D_In[d, :, :, k])

        # Second contraction along the jN_In axis (axis 1 of Vdm, axis 2 of x3D_Buf1)
        buf2 = np.zeros((dim1, n_Out, n_Out, n_In), dtype=x3D_In.dtype)
        for d in range(dim1):
            for k in range(n_Out):
                buf2[d, k, :, :] = Vdm @ np.ascontiguousarray(buf1[d, k, :, :])

        # Third contraction along the kN_In axis (axis 1 of Vdm, axis 3 of x3D_Buf2)
        buf3_Flat = buf2.reshape(-1, n_In) @ Vdm.T

        return buf3_Flat.reshape(dim1, n_Out, n_Out, n_Out)


if not NUMBA_AVAILABLE:
    def change_basis_2D(Vdm: np.ndarray, x2D_In: np.ndarray) -> np.ndarray:
        """ Interpolate a 2D tensor product Lagrange basis defined by (N_in+1) 1D interpolation point positions xi_In(0:N_In)
            to another 2D tensor product node positions (number of nodes N_out+1)
            defined by (N_out+1) interpolation point positions xi_Out(0:N_Out)
            xi is defined in the 1D reference element xi=[-1,1]
        """
        # INFO: numpy-version, not compatible with numba
        # First contraction along the iN_In axis (axis 1 of Vdm, axis 1 of x2D_In)
        # x2D_Buf1 = np.tensordot(Vdm, x2D_In, axes=(1, 1))
        # x2D_Buf1 = np.moveaxis(x2D_Buf1, 0, 1)  # Correct the shape to (dim1, n_Out, n_In, n_In)
        x2D_Buf1 = Vdm @ x2D_In

        # Second contraction along the jN_In axis (axis 1 of Vdm, axis 2 of x2D_Buf1)
        # x2D_Out = np.tensordot(Vdm, x2D_Buf1, axes=(1, 2))
        # x2D_Out = np.moveaxis(x2D_Out, 0, 2)  # Correct the shape to  (dim1, n_Out, n_Out, n_In)
        # x2D_Out = x2D_Buf1 @ Vdm.T
        # PERF: This is actually slower than the individual contractions
        # x2D_Out = np.einsum('pi,qj,dij->dpq', Vdm, Vdm, x2D_In, optimize=True)

        return x2D_Buf1 @ Vdm.T

else:
    @jit((types.float64[:, :, ::1])(types.float64[:, ::1], types.float64[:, :, :]), nopython=True, cache=True, nogil=True)
    def change_basis_2D(Vdm: np.ndarray, x2D_In: np.ndarray) -> np.ndarray:
        """ Interpolate a 2D tensor product Lagrange basis defined by (N_in+1) 1D interpolation point positions xi_In(0:N_In)
            to another 2D tensor product node positions (number of nodes N_out+1)
            defined by (N_out+1) interpolation point positions xi_Out(0:N_Out)
            xi is defined in the 1D reference element xi=[-1,1]
        """
        dim1, n_In, _ = x2D_In.shape
        n_Out = Vdm.shape[0]

        # First contraction along the iN_In axis (axis 1 of Vdm, axis 1 of x2D_In)
        x2D_Flat = np.ascontiguousarray(x2D_In)
        buf1 = np.zeros((dim1, n_Out, n_In), dtype=x2D_In.dtype)
        for d in range(dim1):
            buf1[d] = Vdm @ x2D_Flat[d]

        # Second contraction along the jN_In axis (axis 1 of Vdm, axis 2 of buf1)
        # Result shape: (dim1, n_Out, n_Out)
        # Reshaping to 2D for a single GEMM call: (dim1 * n_Out, n_In) @ (n_In, n_Out)
        x2D_Out = buf1.reshape(-1, n_In) @ Vdm.T

        return x2D_Out.reshape(dim1, n_Out, n_Out)


def evaluate_jacobian(xGeo_In: np.ndarray, VdmGLtoAP: np.ndarray, D_EqToGL: np.ndarray) -> np.ndarray:
    """ Calculate the Jacobian of the mapping for a given element
    """
    dim1  = xGeo_In.shape[0]
    n_In  = xGeo_In.shape[1]
    n_Out = D_EqToGL.shape[0]

    # Perform tensor contraction for the first derivative (Xi direction)
    # dXdXiGL   = np.tensordot(D_EqToGL, xGeo_In, axes=(1, 1))
    # dXdXiGL   = np.moveaxis(dXdXiGL  , 1, 0)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)
    _X1     = xGeo_In.reshape(dim1, n_In, n_In * n_In)
    dXdXiGL = (D_EqToGL @ _X1).reshape(dim1, n_Out, n_In, n_In)

    # Perform tensor contraction for the second derivative (Eta direction)
    dXdEtaGL  = np.tensordot(D_EqToGL, xGeo_In, axes=(1, 2))
    dXdEtaGL  = np.moveaxis(dXdEtaGL , 1, 0)  # Correct the shape to (3, n_Out, n_In, n_In)
    # PERF: This is actually slower than the individual contractions
    # dXdEtaGL  = np.einsum('qj,dijk->dqik', D_EqToGL, xGeo_In, optimize=True)

    # Perform tensor contraction for the third derivative (Zeta direction)
    # dXdZetaGL = np.tensordot(D_EqToGL, xGeo_In, axes=(1, 3))
    # dXdZetaGL = np.moveaxis(dXdZetaGL, 1, 0)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)
    _X3       = xGeo_In.reshape(dim1, n_In * n_In, n_In)
    dXdZetaGL = (_X3 @ D_EqToGL.T).reshape(dim1, n_In, n_In, n_Out)
    dXdZetaGL = np.transpose(dXdZetaGL, (0, 3, 1, 2))
    # PERF: This is actually slower than the individual contractions
    # dXdZetaGL = np.einsum('rk,dijk->drij', D_EqToGL, xGeo_In, optimize=True)

    # Change basis for each direction
    dXdXiAP   = change_basis_3D(VdmGLtoAP, dXdXiGL  )
    dXdEtaAP  = change_basis_3D(VdmGLtoAP, dXdEtaGL )
    dXdZetaAP = change_basis_3D(VdmGLtoAP, dXdZetaGL)

    # Precompute cross products between dXdEtaAP and dXdZetaAP for all points
    # cross_eta_zeta = np.cross(dXdEtaAP, dXdZetaAP, axis=0)  # Shape: (3, nGeoRef, nGeoRef, nGeoRef)
    # > Manually compute cross product
    cross_eta_zeta = np.empty_like(dXdEtaAP)
    cross_eta_zeta[0] = dXdEtaAP[1] * dXdZetaAP[2] - dXdEtaAP[2] * dXdZetaAP[1]
    cross_eta_zeta[1] = dXdEtaAP[2] * dXdZetaAP[0] - dXdEtaAP[0] * dXdZetaAP[2]
    cross_eta_zeta[2] = dXdEtaAP[0] * dXdZetaAP[1] - dXdEtaAP[1] * dXdZetaAP[0]

    # Fill output Jacobian array
    # jacOut = np.einsum('ijkl,ijkl->jkl', dXdXiAP, cross_eta_zeta)
    # PERF: This is actually slower than the individual contractions
    # jacOut = np.sum(dXdXiAP * cross_eta_zeta, axis=0)

    return np.einsum('ijkl,ijkl->jkl', dXdXiAP, cross_eta_zeta)


def evaluate_jacobian_simplex(xGeo_In: np.ndarray, _: np.ndarray, D_EqToGL: np.ndarray) -> np.ndarray:
    # Perform tensor contraction for each derivative
    # Change basis for each direction
    dXdXiAP   = xGeo_In[:, :] @ D_EqToGL[0, :, :]
    dXdEtaAP  = xGeo_In[:, :] @ D_EqToGL[1, :, :]
    dXdZetaAP = xGeo_In[:, :] @ D_EqToGL[2, :, :]

    # Manually compute cross product
    cross_eta_zeta = np.empty_like(dXdEtaAP)
    cross_eta_zeta[0] = dXdEtaAP[1] * dXdZetaAP[2] - dXdEtaAP[2] * dXdZetaAP[1]
    cross_eta_zeta[1] = dXdEtaAP[2] * dXdZetaAP[0] - dXdEtaAP[0] * dXdZetaAP[2]
    cross_eta_zeta[2] = dXdEtaAP[0] * dXdZetaAP[1] - dXdEtaAP[1] * dXdZetaAP[0]

    # Fill output Jacobian array
    # jacOut = np.sum(dXdXiAP * cross_eta_zeta, axis=0)

    return np.sum(dXdXiAP * cross_eta_zeta, axis=0)

# INFO: ALTERNATIVE VERSION, CACHING VDM, D
# class JacobianEvaluator:
#     def __init__(self, VdmGLtoAP: np.ndarray, D_EqToGL: np.ndarray) -> None:
#         self.VdmGLtoAP: Final[np.ndarray] = VdmGLtoAP
#         self.D_EqToGL:  Final[np.ndarray] = D_EqToGL
#
#     def evaluate_jacobian(self, xGeo_In: np.ndarray) -> np.ndarray:
#         # Perform tensor contraction for the first derivative (Xi direction)
#         dXdXiGL   = np.tensordot(self.D_EqToGL, xGeo_In, axes=(1, 1))
#         dXdXiGL   = np.moveaxis(dXdXiGL  , 1, 0)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)
#
#         # Perform tensor contraction for the second derivative (Eta direction)
#         dXdEtaGL  = np.tensordot(self.D_EqToGL, xGeo_In, axes=(1, 2))
#         dXdEtaGL  = np.moveaxis(dXdEtaGL , 1, 0)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)
#
#         # Perform tensor contraction for the third derivative (Zeta direction)
#         dXdZetaGL = np.tensordot(self.D_EqToGL, xGeo_In, axes=(1, 3))
#         dXdZetaGL = np.moveaxis(dXdZetaGL, 1, 0)  # Correct the shape to (3, nGeoRef, nGeoRef, nGeoRef)
#
#         # Change basis for each direction
#         dXdXiAP   = change_basis_3D(self.VdmGLtoAP, dXdXiGL  )
#         dXdEtaAP  = change_basis_3D(self.VdmGLtoAP, dXdEtaGL )
#         dXdZetaAP = change_basis_3D(self.VdmGLtoAP, dXdZetaGL)
#
#         # Precompute cross products between dXdEtaAP and dXdZetaAP for all points
#         cross_eta_zeta = np.cross(dXdEtaAP, dXdZetaAP, axis=0)  # Shape: (3, nGeoRef, nGeoRef, nGeoRef)
#
#         # Fill output Jacobian array
#         jacOut = np.einsum('ijkl,ijkl->jkl', dXdXiAP, cross_eta_zeta)
#
#         return jacOut



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
# Copyright (c) 2022 Gabriel Altay (Original Version)
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
from typing import List, Literal, Iterable, Union, overload
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import numpy as np
import numpy.typing as npt
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
# ==================================================================================================================================


def HilbertCurveNumpy() -> None:
    """ Monkey-patch for hilbertcurve.HilbertCurve:
        - Adds a NumPy-vectorized _distances_from_points_numpy(self, points)
        - Wraps distances_from_points to use the vectorized path for ndarray inputs

        Assumptions:
        - points is (M, n) with integer coordinates in [0, 2**p - 1]
        - p * n <= 63 so distances fit into uint64

        The patch is idempotent: calling apply_hilbert_numpy_patch() multiple times is harmless
    """
    # If the package isn't available, silently skip patching
    try:
        from hilbertcurve.hilbertcurve import HilbertCurve
    except Exception:
        return None

    # Avoid re-patching
    if getattr(HilbertCurve, '_numpy_patch_applied', False):
        return None

    # Typing helpers
    @overload
    def _distances_from_points_numpy(self, points: npt.NDArray, match_type: Literal[False] = False) -> List: ...
    @overload
    def _distances_from_points_numpy(self, points: List       , match_type: Literal[True])          -> List: ...
    @overload
    def _distances_from_points_numpy(self, points: npt.NDArray, match_type: Literal[True])          -> npt.NDArray: ...
    # Function
    def _distances_from_points_numpy(self,
                                     points    : Union[List, np.ndarray],
                                     match_type: bool = False) -> Union[np.ndarray, list]:
        """ Batch implementation for distances_from_points in numpy
        """
        pts = np.asarray(points)
        if pts.ndim != 2 or pts.shape[1] != self.n:
            raise ValueError(f'points must be (M, {self.n})')

        # Work in uint64 for per-coordinate logic (coordinates fit in 64-bit)
        # > Copy to not mutate caller memory
        upts = pts.astype(np.uint64, copy=True)

        # Range checks
        max_x = np.uint64(self.max_x)
        if (upts > max_x).any():
            raise ValueError('point coordinates out of range [0, 2**p - 1]')

        M     = upts.shape[0]
        n     = self.n
        pbits = self.p

        # Inverse undo excess work
        q = np.uint64(1) << np.uint64(pbits - 1)  # m
        while q > 1:
            pmask = q - 1
            for i in range(n):
                mask = (upts[:, i] & q) != 0
                if mask.any():
                    upts[mask, 0] ^= pmask
                if (~mask).any():
                    t = (upts[~mask, 0] ^ upts[~mask, i]) & pmask
                    upts[~mask, 0] ^= t
                    upts[~mask, i] ^= t
            q >>= 1

        # Gray encode
        for i in range(1, n):
            upts[:, i] ^= upts[:, i - 1]

        tmask = np.zeros(M, dtype=np.uint64)
        q = np.uint64(1) << np.uint64(pbits - 1)
        while q > 1:
            mask = (upts[:, n - 1] & q) != 0
            if mask.any():
                tmask[mask] ^= (q - 1)
            q >>= 1
        upts ^= tmask[:, None]

        # Interleave bit-planes into a big (potentially >64-bit) integer per row
        # > Build as Python ints to avoid overflow; keep it efficient by
        # >   - computing the per-plane "combine" (0..(2^n-1)) vectorized,
        # >   - updating the big integer with one vectorized object op per plane
        # > Using dtype=object leverages Python big-int shifts/ors elementwise
        h = np.zeros(M, dtype=object)

        # Precompute weights across dims for each plane: (1 << (n-1-i))
        dim_weights = np.array([1 << (n - 1 - i) for i in range(n)], dtype=np.uint64)

        for bit in range(pbits - 1, -1, -1):
            # bits_plane: (M, n) of 0/1 for current bit
            bits_plane = (upts >> np.uint64(bit)) & np.uint64(1)
            # Combine per row into [0 .. 2^n-1] by weighted sum across dims
            combine = (bits_plane * dim_weights).sum(axis=1).astype(np.uint64)
            # Update big-int index: shift by n, then OR the small combine
            # > Cast combine to object just for the OR
            h = (h << n) | combine.astype(object)

        # Result type parity
        if match_type and isinstance(points, np.ndarray):
            return np.array(h, dtype=points.dtype, copy=False)

        return list(map(int, h))

    # Attach the vectorized helper
    HilbertCurve._distances_from_points_numpy = _distances_from_points_numpy  # type: ignore[attr-defined]

    # Wrap the public API to prefer the NumPy path
    _orig_dfp = HilbertCurve.distances_from_points

    def _dfp_patched(self, points: Iterable[Iterable[int]], match_type: bool = False):
        """ Wrapper for the monkey-patched numby implementation
        """
        # Prefer NumPy path when possible
        if isinstance(points, np.ndarray):
            try:
                distances = self._distances_from_points_numpy(points, match_type=match_type)
            except Exception as e:
                raise RuntimeError(f'HilbertCurve.distances_from_points_numpy encountered an unexpected error: {e}')
                # Fallback to original behavior on any unexpected issue
                # distances = _orig_dfp(self, points, match_type=match_type)
        else:
            distances = _orig_dfp(self, points, match_type=match_type)

        return distances

    HilbertCurve.distances_from_points = _dfp_patched
    HilbertCurve._numpy_patch_applied  = True  # type: ignore[attr-defined]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This file is part of PyHOPE
#
# Copyright (c) 2024 Numerics Research Group, University of Stuttgart, Prof. Andrea Beck
#
# PyHOPE is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# PyHOPE is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# PyHOPE. If not, see <http://www.gnu.org/licenses/>.

# ==================================================================================================================================
# Mesh generation library
# ==================================================================================================================================
# ----------------------------------------------------------------------------------------------------------------------------------
# Standard libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import os
from argparse import Namespace
from typing import Optional, final
# ----------------------------------------------------------------------------------------------------------------------------------
# Third-party libraries
# ----------------------------------------------------------------------------------------------------------------------------------
import argparse
# ----------------------------------------------------------------------------------------------------------------------------------
# Local imports
# ----------------------------------------------------------------------------------------------------------------------------------
# ----------------------------------------------------------------------------------------------------------------------------------
# Local definitions
# ----------------------------------------------------------------------------------------------------------------------------------
STD_LENGTH = 79  # Standard length for output to console
PAR_LENGTH = 32
DEF_LENGTH = 20
# ==================================================================================================================================


def info(string: str, newline: bool = False) -> None:
    if newline:
        print('\n!', string)
    else:
        print('!', string)


def separator(length: int = STD_LENGTH) -> str:
    return '!' + '-'*(length-1)


@final
class CommandLine:
    """ Parse command line arguments, both explicit [*.ini] and flags [--]
    """
    def __init__(self, argv, name: str, version: str, commit: Optional[str]) -> None:
        # Local imports ----------------------------------------
        import pyhope.config.config as config
        from pyhope.output.output import Colors
        # ------------------------------------------------------

        # Read the command line arguments and store everything
        self.argv    = argv
        self.name    = name
        self.version = version
        self.commit  = commit
        # self.help    = ''

        # Print the header
        self.help = (Colors.BANNERA + '!' + '='*(STD_LENGTH-1))
        self.helpjoin(f'! {name} version {version}' + (f' [commit {commit}]' if commit else ''))
        self.helpjoin(Colors.BANNERA + '!' + '='*(STD_LENGTH-1) + Colors.END)

        # Assemble the help output
        for key in config.prms:
            # Check if we encountered a section
            if config.prms[key]['type'] == 'section':
                self.helpjoin(separator())
                self.helpjoin('! {}'.format(key))
                self.helpjoin(separator())
                continue

            if config.prms[key]['default'] is not None:
                default = config.prms[key]['default']
            else:
                default = ''

            # Convert booleans to strings
            if isinstance(default, bool):
                default = 'T' if default else 'F'

            if config.prms[key]['help']:
                help    = config.prms[key]['help']
            else:
                help    = ''

            self.helpjoin(f'{key:<{PAR_LENGTH}} = {default:>{DEF_LENGTH}} ! {help}')

        return None

    def __enter__(self) -> tuple[Namespace, list]:
        # Setup an argument parser and add know arguments
        _ = parser = argparse.ArgumentParser(prog   = self.name,                              # noqa: E251
                                             epilog = self.help,                              # noqa: E251
                                             usage  ='PyHOPE [-h] [-V] ' +                    # noqa: E251
                                                     '[--verify[-install] [tutorials] | --verify-health] ' +
                                                     '[<parameter.ini / mesh.h5>]',
                                             formatter_class=argparse.RawDescriptionHelpFormatter)

        _ = parser.add_argument('-V', '--version',
                                action = 'store_true',                                        # noqa: E251
                                help   = 'display the version number and exit')               # noqa: E251

        verifyParser = parser.add_argument_group('Verification options')
        _ = verifyParser.add_argument('--verify',
                                      nargs   = '?',                                          # noqa: E251
                                      const   = True,                                         # noqa: E251
                                      metavar = 'tutorials',                                  # noqa: E251
                                      help    = 'verify the installation and exit')           # noqa: E251
        _ = verifyParser.add_argument('--verify-health',
                                      action  = 'store_true',                                 # noqa: E251
                                      help    = 'check health and (optional) dependencies')   # noqa: E251
        _ = verifyParser.add_argument('--verify-install',
                                      nargs   = '?',                                          # noqa: E251
                                      const   = True,                                         # noqa: E251
                                      metavar = 'tutorials',                                  # noqa: E251
                                      help    = 'verify the installation and exit')           # noqa: E251
        _ = verifyParser.add_argument('--skip-checks',
                                      action  = 'store_true',                                 # noqa: E251
                                      # Hidden: Disable checks for verification
                                      help    = argparse.SUPPRESS)                            # noqa: E251
        _ = parser.add_argument('input',
                                nargs   = '?',                                                # noqa: E251
                                metavar = '<parameter.ini / mesh.h5>',                        # noqa: E251
                                help    = 'PyHOPE parameter or mesh file')                    # noqa: E251

        # Parse known arguments and return other flags for further processing
        args, argv = parser.parse_known_args()
        return args, argv

    def __exit__(self, *args: object) -> None:
        return None

    def helpjoin(self, end) -> None:
        self.help = os.linesep.join([self.help, end])
